# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
from typing import Dict

from ..util import MultilineFormatter
from .downloaders import IDatasetDownloader, KaggleDatasetDownloader, TfdsDatasetDownloader


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Download a publicly available dataset",
        description=f"""
        Downloads a publicly available dataset and saves it in a given format.|n
        |n
        To download the dataset, run "datum download <dataset_type> get". On the other hand,
        for information about the datasets, run "datum download <dataset_type> describe".
        Supported dataset types are: {list(DOWNLOADERS.keys())}
        """,
        formatter_class=MultilineFormatter,
    )
    subparsers = parser.add_subparsers(title="Dataset types")
    for name, downloader in DOWNLOADERS.items():
        dataset_type_parser = subparsers.add_parser(
            name=name,
            help=f"Download {name} dataset",
            formatter_class=MultilineFormatter,
        )
        _subparsers = dataset_type_parser.add_subparsers(title="Commands")
        build_get_subparser(_subparsers, name, downloader)
        build_describe_subparser(_subparsers, name, downloader)


def build_get_subparser(
    subparsers: argparse._SubParsersAction, name: str, downloader: IDatasetDownloader
):
    parser = subparsers.add_parser(
        name="get",
        help="Download a publicly available dataset",
        description=downloader.get_command_description(),
        formatter_class=MultilineFormatter,
    )
    parser.add_argument("-i", "--dataset-id", required=True, help="Which dataset to download")
    parser.add_argument(
        "-f", "--output-format", help="Output format (default: original format of the dataset)"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument("-s", "--subset", help="Save only the specified subset")
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for output format (pass '-- -h' for help). "
        "Must be specified after the main command arguments",
    )

    parser.set_defaults(command=download_command, downloader=downloader)

    return parser


def build_describe_subparser(
    subparsers: argparse._SubParsersAction, name: str, downloader: IDatasetDownloader
):
    parser = subparsers.add_parser(
        name="describe",
        help="Print information about downloadable datasets",
        description=f"""
        Reports information about datasets that can be downloaded with the
        "datum download {name}" command. The information is reported either as
        human-readable text (the default) or as a JSON object."""
        + downloader.describe_command_description(),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "--report-format",
        choices=("text", "json"),
        default="text",
        help="Format in which to report the information (default: %(default)s)",
    )
    parser.add_argument(
        "--report-file", help="File to which to write the report (default: standard output)"
    )
    parser.set_defaults(command=describe_downloads_command, downloader=downloader)

    return parser


def get_sensitive_args():
    return {
        download_command: ["dst_dir", "extra_args"],
        describe_downloads_command: ["report-file"],
    }


DOWNLOADERS: Dict[str, IDatasetDownloader] = {
    "tfds": TfdsDatasetDownloader,
    "kaggle": KaggleDatasetDownloader,
}


def download_command(args):
    args.downloader.download(
        args.dataset_id,
        args.dst_dir,
        args.overwrite,
        args.output_format,
        args.subset,
        args.extra_args,
    )


def describe_downloads_command(args):
    return args.downloader.describe(args.report_format, args.report_file)



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import csv
import errno
import os.path as osp
from typing import List, Optional

from datumaro.components.annotation import AnnotationType, Bbox, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class VottCsvPath:
    ANNO_FILE_SUFFIX = "-export.csv"


class VottCsvBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find annotations file", path)

        if not subset:
            subset = osp.splitext(osp.basename(path))[0].rsplit("-", maxsplit=1)[0]

        super().__init__(subset=subset, ctx=ctx)

        if has_meta_file(path):
            self._categories = {
                AnnotationType.label: LabelCategories.from_iterable(parse_meta_file(path).keys())
            }
        else:
            self._categories = {AnnotationType.label: LabelCategories()}

        self._items = list(self._load_items(path).values())

    def _load_items(self, path):
        items = {}

        label_categories = self._categories[AnnotationType.label]

        with open(path, encoding="utf-8") as content:
            for row in csv.DictReader(content):
                item_id = osp.splitext(row["image"])[0]

                if item_id not in items:
                    items[item_id] = DatasetItem(
                        id=item_id,
                        subset=self._subset,
                        media=Image.from_file(path=osp.join(osp.dirname(path), row["image"])),
                    )

                annotations = items[item_id].annotations

                label_name = row.get("label")
                x_min = row.get("xmin")
                y_min = row.get("ymin")
                x_max = row.get("xmax")
                y_max = row.get("ymax")

                if label_name and x_min and y_min and x_max and y_max:
                    label_idx = label_categories.find(label_name)[0]
                    if label_idx is None:
                        label_idx = label_categories.add(label_name)

                    x_min = float(x_min)
                    y_min = float(y_min)
                    w = float(x_max) - x_min
                    h = float(y_max) - y_min

                    annotations.append(Bbox(x_min, y_min, w, h, label=label_idx))
                    self._ann_types.add(AnnotationType.bbox)

        return items


class VottCsvImporter(Importer):
    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, ".csv", "vott_csv")

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file("*" + VottCsvPath.ANNO_FILE_SUFFIX)

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [osp.splitext(VottCsvPath.ANNO_FILE_SUFFIX)[1]]



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from typing import Sequence

from datumaro.components.contexts.importer import _ImportFail
from datumaro.components.dataset_base import SubsetBase
from datumaro.components.merge import ExactMerge
from datumaro.components.merge.extractor_merger import ExtractorMerger, check_identicalness
from datumaro.plugins.data_formats.coco.base import _CocoBase


class COCOTaskMergedBase(SubsetBase):
    __not_plugin__ = True

    def __init__(
        self,
        sources: Sequence[_CocoBase],
        subset: str,
    ):
        super().__init__(subset=subset, ctx=None)
        self._infos = check_identicalness([s.infos() for s in sources])
        self._media_type = check_identicalness([s.media_type() for s in sources])
        self._is_stream = check_identicalness([s.is_stream for s in sources])
        self._categories = ExactMerge.merge_categories([s.categories() for s in sources])

        self._sources = sources
        self._item_keys = None

    def __iter__(self):
        if len(self._sources) == 1:
            yield from self._sources[0]
        else:
            for item_key in self.item_keys:
                items = [
                    item
                    for s in self._sources
                    if (item := s.get_dataset_item(item_key)) is not None
                ]
                assert len(items) > 0

                item, remainders = items[0], items[1:]

                for remainder in remainders:
                    item = ExactMerge.merge_items(item, remainder)
                yield item

    def __len__(self):
        if len(self._sources) == 1:
            return len(self._sources[0])
        else:
            return len(self.item_keys)

    @property
    def item_keys(self):
        if self._item_keys is None:
            self._item_keys = set()
            for s in self._sources:
                self._item_keys.update(s.iter_item_ids())

        return self._item_keys

    @property
    def is_stream(self) -> bool:
        return self._is_stream


class COCOExtractorMerger(ExtractorMerger):
    __not_plugin__ = True

    def __init__(self, sources: Sequence[_CocoBase]):
        if len(sources) == 0:
            raise _ImportFail("It should not be empty.")

        self._infos = check_identicalness([s.infos() for s in sources])
        self._media_type = check_identicalness([s.media_type() for s in sources])
        self._is_stream = check_identicalness([s.is_stream for s in sources])
        self._categories = ExactMerge.merge_categories([s.categories() for s in sources])

        ann_types = set()
        for source in sources:
            ann_types.union(source.ann_types())
        self._ann_types = ann_types

        grouped_by_subset = defaultdict(list)

        for s in sources:
            grouped_by_subset[s.subset] += [s]

        self._subsets = {
            subset: [COCOTaskMergedBase(sources, subset)]
            for subset, sources in grouped_by_subset.items()
        }



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Tuple

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset_base import IDataset
from datumaro.errors import DatasetError

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from pandas import DataFrame, Series


__all__ = ["LossDynamicsAnalyzer", "NoisyLabelCandidate"]


@dataclass(order=True, frozen=True)
class NoisyLabelCandidate:
    id: str = field(compare=False)
    subset: str = field(compare=False)
    ann_id: int = field(compare=False)
    label_id: int = field(compare=False)
    metric: float = field(compare=True)

    def __eq__(self, __o: "NoisyLabelCandidate") -> bool:
        return self.id == __o.id and self.subset == __o.subset and self.ann_id == __o.ann_id


class LossDynamicsAnalyzer:
    """A class for analyzing the dynamics of training loss to identify noisy labels.

    This class parses the dataset to extract information about the training loss dynamics.
    It then calculates the exponential moving average (EMA) of the training loss dynamics.
    A higher EMA value of training loss dynamics can indicate a noisy labeled sample [1]_.
    The class provides an interface to extract the top-k candidates for noisy labels
    based on the statistics. Additionally, it can plot the EMA curves of loss dynamics for the candidates,
    allowing comparison of the dataset's overall average or averages grouped by labels.

    .. [1] Zhou, Tianyi, Shengjie Wang, and Jeff Bilmes.
    "Robust curriculum learning: from clean label detection to noisy label self-correction."
    International Conference on Learning Representations. 2021.
    """

    allowed_task_names = {"OTX-MultiClassCls", "OTX-Det"}
    allowed_tracking_loss_types = {"cls", "bbox", "centerness", "bbox_refine"}

    def __init__(
        self, dataset: IDataset, alpha: float = 0.001, tracking_loss_type: Optional[str] = None
    ) -> None:
        purpose = dataset.infos().get("purpose")
        if purpose != "noisy_label_detection":
            raise DatasetError(
                f'Dataset infos should have purpose="noisy_label_detection", but it has {purpose}.'
            )

        task = dataset.infos().get("task")
        if task not in self.allowed_task_names:
            raise DatasetError(
                f"task={task} is not allowed. The allowed task names are {self.allowed_task_names}."
            )
        elif task == "OTX-Det" and tracking_loss_type is None:
            raise DatasetError(
                'If task="OTX-Det", tracking_loss_type should not be None. '
                f"The allowed tracking loss types are {self.allowed_tracking_loss_types}."
            )
        elif (
            tracking_loss_type is not None
            and tracking_loss_type not in self.allowed_tracking_loss_types
        ):
            raise DatasetError(
                f"tracking_loss_type={tracking_loss_type} "
                f"but the allowed tracking loss types are {self.allowed_tracking_loss_types}."
            )

        self._dataset = dataset
        self._alpha = alpha
        self._df = self._parse_to_dataframe(dataset, alpha, tracking_loss_type)
        self._mean_loss_dyns = self._df.mean(0)
        self._mean_loss_dyns_per_label = {
            label_id: df_sub.mean(0) for label_id, df_sub in self._df.groupby("ann_label")
        }
        self._tracking_loss_type = tracking_loss_type

    @property
    def alpha(self) -> float:
        """A parameter to obtain EMA loss dynamics statistics.

        ema_loss_dyns(t) := (1 - alpha) * ema_loss_dyns(t - 1) + alpha * loss_dyns(t)"""
        return self._alpha

    @property
    def mean_loss_dyns(self) -> Series:
        """Pandas Series object obtained by averaging all EMA loss dynamics statistics"""
        return self._mean_loss_dyns

    @property
    def mean_loss_dyns_per_label(self) -> Dict[LabelCategories.Category, Series]:
        """A dictionary of Pandas Series object obtained
        by averaging EMA loss dynamics statistics according to the label category"""
        label_categories = self._dataset.categories()[AnnotationType.label]
        return {label_categories[k]: v for k, v in self._mean_loss_dyns_per_label.items()}

    @property
    def ema_dataframe(self) -> DataFrame:
        """Pandas DataFrame including full EMA loss dynamics statistics."""
        return self._df

    @staticmethod
    def _parse_to_dataframe(
        dataset: IDataset, ema_alpha: float = 0.001, tracking_loss_type: Optional[str] = None
    ) -> DataFrame:
        """Parse loss dynamics statistics from Datumaro dataset to Pandas DataFrame."""
        key = (
            "loss_dynamics" if tracking_loss_type is None else f"loss_dynamics_{tracking_loss_type}"
        )
        import pandas as pd

        ema_loss_dyns_list = []
        for item in dataset:
            for ann in item.annotations:
                # The first value should be start from zero
                loss_dyns = pd.Series(
                    [0.0] + ann.attributes.get(key),
                    index=[-1] + ann.attributes.get("iters"),
                    name=(item.id, item.subset, ann.id, ann.label),
                )
                ema_loss_dyns = loss_dyns.ewm(alpha=ema_alpha, adjust=False).mean().iloc[1:]
                ema_loss_dyns_list.append(ema_loss_dyns)

        df = pd.DataFrame(ema_loss_dyns_list).rename_axis(
            ["item_id", "item_subset", "ann_id", "ann_label"]
        )
        df.sort_index(axis=1, inplace=True)
        df["last_value"] = df.apply(lambda row: row[row.last_valid_index()], axis=1)
        df.insert(0, "last_value", df.pop("last_value"))
        df.sort_values(by="last_value", inplace=True)

        return df

    def get_top_k_cands(self, top_k: int) -> List[NoisyLabelCandidate]:
        """Return a list of top-k noisy label candidates.

        Parameters
        ----------
        top_k: int
            An integer value to determine the number of candidates

        Returns
        -------
            A list of top-k noisy label candidates.
            It is sorted in descending order by the last value of the EMA training loss dynamics.
        """
        return [
            NoisyLabelCandidate(item_id, item_subset, ann_id, label_id, value)
            for (item_id, item_subset, ann_id, label_id), value in self._df["last_value"]
            .iloc[-top_k:]
            .items()
        ][::-1]

    def plot_ema_loss_dynamics(
        self,
        candidates: Sequence[NoisyLabelCandidate],
        mode: str = "mean",
        mean_plot_style: str = "--",
        mean_plot_color: str = "k",
        figsize: Tuple[int, int] = (4, 3),
        **kwargs,
    ) -> Figure:
        import matplotlib.pyplot as plt

        if mode == "mean":
            cands_by_label_id = {None: candidates}
        elif mode == "label_mean":
            cands_by_label_id = defaultdict(list)
            for cand in candidates:
                cands_by_label_id[cand.label_id].append(cand)
        else:
            raise ValueError(f'mode={mode} is not allowed, it should be "mean" or "label_mean".')

        label_categories = self._dataset.categories()[AnnotationType.label]
        n_cols = len(cands_by_label_id)
        fig, _ = plt.subplots(1, n_cols, figsize=(n_cols * figsize[0], figsize[1]))

        for idx, label_id in enumerate(sorted(list(cands_by_label_id.keys()))):
            plt.subplot(1, n_cols, idx + 1)

            # We have to drop 0 column since it is "last_value" column.
            mean_series = (
                self._mean_loss_dyns.iloc[1:]
                if label_id is None
                else self._mean_loss_dyns_per_label[label_id].iloc[1:]
            )
            plt.plot(mean_series, color=mean_plot_color, linestyle=mean_plot_style, label="mean")

            cands = cands_by_label_id[label_id]
            for cand in cands:
                row = self._df.loc[(cand.id, cand.subset, cand.ann_id, cand.label_id)]
                plt.plot(
                    row.dropna().iloc[1:],
                    label=f"id={cand.id}, subset={cand.subset}, ann_id={cand.ann_id}, label_id={cand.label_id}",
                    **kwargs,
                )

            plt.ylabel("EMA training loss dynamics")
            plt.xlabel("Training iterations")

            plt.title(
                f"Compare the candidates with the average of ({label_categories[label_id].name})"
                if label_id is not None
                else "Compare the candidates with the average of dataset"
            )
            plt.legend(
                loc="upper center", bbox_to_anchor=(0.5, -0.05), fancybox=True, shadow=True, ncol=1
            )

        return fig



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os.path as osp
from collections import OrderedDict
from glob import glob
from typing import Optional

import numpy as np

from datumaro.components.annotation import (
    AnnotationType,
    ExtractedMask,
    LabelCategories,
    MaskCategories,
)
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import InvalidAnnotationError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image
from datumaro.util.image import find_images
from datumaro.util.mask_tools import generate_colormap, load_mask
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file

from .format import (
    SynthiaAlLabelMap,
    SynthiaAlPath,
    SynthiaRandLabelMap,
    SynthiaRandPath,
    SynthiaSfLabelMap,
    SynthiaSfPath,
)


def make_categories(label_map):
    categories = {}
    label_categories = LabelCategories()
    for label in label_map:
        label_categories.add(label)
    categories[AnnotationType.label] = label_categories

    has_colors = any(v is not None for v in label_map.values())
    if not has_colors:  # generate new colors
        colormap = generate_colormap(len(label_map))
    else:  # only copy defined colors
        colormap = {
            label_id: (desc[0], desc[1], desc[2])
            for label_id, desc in enumerate(label_map.values())
        }
    mask_categories = MaskCategories(colormap)
    mask_categories.inverse_colormap  # pylint: disable=pointless-statement
    categories[AnnotationType.mask] = mask_categories
    return categories


def parse_label_map(path):
    label_map = OrderedDict()
    with open(path, "r", encoding="utf-8") as f:
        for line in f:
            # skip empty and commented lines
            line = line.strip()
            if not line or line[0] == "#":
                continue

            # color, name
            label_desc = line.split()

            if 2 < len(label_desc):
                name = label_desc[3]
                color = tuple([int(c) for c in label_desc[:3]])
            else:
                name = label_desc[0]
                color = None

            if name in label_map:
                raise InvalidAnnotationError("Label '%s' is already defined" % name)

            label_map[name] = color
    return label_map


class _SynthiaBase(SubsetBase):
    def __init__(
        self,
        path: str,
        path_formats,
        label_map,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        super().__init__(subset=subset, ctx=ctx)

        self._path_formats = path_formats
        self._label_map = label_map
        self._ann_types = set()

        self._img_dir = None
        self._inst_dir = None
        self._seg_dir = None
        for path_format in vars(path_formats).keys():
            if path_format == "IMAGES_DIR":
                self._img_dir = osp.join(path, path_formats.IMAGES_DIR)
            elif path_format == "LABELS_SEGM_DIR":
                self._inst_dir = osp.join(path, path_formats.LABELS_SEGM_DIR)
            elif path_format == "SEMANTIC_SEGM_DIR":
                self._seg_dir = osp.join(path, path_formats.SEMANTIC_SEGM_DIR)

        self._categories = self._load_categories(path)
        self._items = list(self._load_items().values())

    def _load_categories(self, path):
        if has_meta_file(path):
            return make_categories(parse_meta_file(path))

        label_map_path = osp.join(path, "label_colors.txt")
        if osp.isfile(label_map_path):
            label_map = parse_label_map(label_map_path)
        else:
            label_map = self._label_map

        return make_categories(label_map)

    def _load_items(self):
        if self._img_dir and osp.isdir(self._img_dir):
            images = {
                osp.splitext(osp.relpath(p, self._img_dir))[0].replace("\\", "/"): p
                for p in find_images(self._img_dir, recursive=True)
            }
        else:
            images = {}

        items = {}
        if self._inst_dir and osp.isdir(self._inst_dir):
            gt_labels = glob(self._inst_dir + "/*.txt")
            for gt_label in gt_labels:
                item_id = osp.splitext(osp.relpath(gt_label, self._inst_dir))[0].replace("\\", "/")

                anno = []
                labels_mask = np.loadtxt(gt_label)
                classes = np.unique(labels_mask)
                for label_id in classes:
                    anno.append(
                        ExtractedMask(index_mask=labels_mask, index=label_id, label=label_id)
                    )
                    self._ann_types.add(AnnotationType.mask)

                image = images.get(item_id)
                if image:
                    image = Image.from_file(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)
        elif self._seg_dir and osp.isdir(self._seg_dir):
            for seg_img_path in find_images(self._seg_dir, recursive=True):
                item_id = osp.splitext(osp.relpath(seg_img_path, self._seg_dir))[0].replace(
                    "\\", "/"
                )

                inverse_cls_colormap = self._categories[AnnotationType.mask].inverse_colormap

                color_mask = load_mask(seg_img_path, inverse_cls_colormap, default_id=0)

                anno = []
                classes = np.unique(color_mask)
                for label_id in classes:
                    anno.append(
                        ExtractedMask(index_mask=color_mask, index=label_id, label=label_id)
                    )
                    self._ann_types.add(AnnotationType.mask)

                image = images.get(item_id)
                if image:
                    image = Image.from_file(path=image)

                items[item_id] = DatasetItem(id=item_id, media=image, annotations=anno)

        return items


class SynthiaRandBase(_SynthiaBase):
    def __init__(self, path: str, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaRandPath, label_map=SynthiaRandLabelMap, **kwargs
        )


class SynthiaSfBase(_SynthiaBase):
    def __init__(self, path, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaSfPath, label_map=SynthiaSfLabelMap, **kwargs
        )


class SynthiaAlBase(_SynthiaBase):
    def __init__(self, path, **kwargs):
        super().__init__(
            path=path, path_formats=SynthiaAlPath, label_map=SynthiaAlLabelMap, **kwargs
        )



# Copyright (C) 2021-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import errno
import io
import os
import os.path as osp
import shutil
from copy import deepcopy
from enum import IntEnum
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import cv2
import imagesize
import numpy as np

from datumaro.components.crypter import NULL_CRYPTER, Crypter
from datumaro.components.errors import DatumaroError, MediaShapeError
from datumaro.util.definitions import BboxIntCoords
from datumaro.util.image import (
    _image_loading_errors,
    copyto_image,
    decode_image,
    lazy_image,
    load_image,
    save_image,
)

if TYPE_CHECKING:
    import pandas as pd
else:
    from datumaro.util.import_util import lazy_import

    pd = lazy_import("pandas")


AnyData = TypeVar("AnyData", bytes, np.ndarray)


class MediaType(IntEnum):
    NONE = 0
    MEDIA_ELEMENT = 1
    IMAGE = 2
    BYTE_IMAGE = 3
    VIDEO_FRAME = 4
    VIDEO = 5
    POINT_CLOUD = 6
    MULTIFRAME_IMAGE = 7
    ROI_IMAGE = 8
    MOSAIC_IMAGE = 9
    TABLE_ROW = 10

    @property
    def media(self) -> Optional[Type[MediaElement]]:
        if self == self.NONE:
            return None
        if self == self.MEDIA_ELEMENT:
            return MediaElement
        if self == self.IMAGE:
            return Image
        if self == self.VIDEO_FRAME:
            return VideoFrame
        if self == self.VIDEO:
            return Video
        if self == self.POINT_CLOUD:
            return PointCloud
        if self == self.MULTIFRAME_IMAGE:
            return MultiframeImage
        if self == self.ROI_IMAGE:
            return RoIImage
        if self == self.MOSAIC_IMAGE:
            return MosaicImage
        if self == self.TABLE_ROW:
            return TableRow
        raise NotImplementedError


class MediaElement(Generic[AnyData]):
    _type = MediaType.MEDIA_ELEMENT

    def __init__(self, crypter: Crypter = NULL_CRYPTER, *args, **kwargs) -> None:
        self._crypter = crypter

    def as_dict(self) -> Dict[str, Any]:
        # NOTE:
        # attributes starting with a single underscore are assumed
        # to be arguments of __init__ method and
        # attributes starting with double underscores are assuemd
        # to be not directly related to __init__ method.
        return {
            key[1:]: value
            for key, value in self.__dict__.items()
            if key.startswith("_") and not key.startswith(f"_{self.__class__.__name__}")
        }

    def from_self(self, **kwargs):
        attrs = deepcopy(self.as_dict())
        attrs.update(kwargs)
        return self.__class__(**attrs)

    @property
    def is_encrypted(self) -> bool:
        return not self._crypter.is_null_crypter

    def set_crypter(self, crypter: Crypter):
        self._crypter = crypter

    @property
    def type(self) -> MediaType:
        return self._type

    @property
    def data(self) -> Optional[AnyData]:
        return None

    @property
    def has_data(self) -> bool:
        return False

    @property
    def bytes(self) -> Optional[bytes]:
        return None

    def __eq__(self, other: object) -> bool:
        other_type = getattr(other, "type", None)
        if self.type != other_type:
            return False
        return True

    def save(
        self,
        fp: Union[str, io.IOBase],
        crypter: Crypter = NULL_CRYPTER,
    ):
        raise NotImplementedError


class FromFileMixin:
    def __init__(self, path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert path, "Path can't be empty"
        self._path = path

    @property
    def path(self) -> str:
        """Path to the media file"""
        # TODO: do we need this replace?
        return self._path.replace("\\", "/")

    @property
    def bytes(self) -> Optional[bytes]:
        if self.has_data:
            with open(self._path, "rb") as f:
                _bytes = f.read()
            return _bytes
        return None

    @property
    def has_data(self) -> bool:
        return os.path.exists(self.path)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={repr(self._path)})"


class FromDataMixin(Generic[AnyData]):
    def __init__(self, data: Union[Callable[[], AnyData], AnyData], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = data

    @property
    def data(self) -> Optional[AnyData]:
        if callable(self._data):
            return self._data()
        return self._data

    @property
    def bytes(self) -> Optional[bytes]:
        if self.has_data:
            _bytes = self._data() if callable(self._data) else self._data
            if isinstance(_bytes, bytes):
                return _bytes
        return None

    @property
    def has_data(self) -> bool:
        return self._data is not None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(data=" + repr(self._data)[:20].replace("\n", "") + "...)"


class Image(MediaElement[np.ndarray]):
    _type = MediaType.IMAGE

    _DEFAULT_EXT = ".png"

    def __init__(
        self,
        size: Optional[Tuple[int, int]] = None,
        ext: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        assert self.__class__ != Image, (
            f"Directly initalizing {self.__class__.__name__} is not supported. "
            f"Please use one of fractory functions ({self.__class__.__name__}.from_file(), "
            f"{self.__class__.__name__}.from_numpy(), {self.__class__.__name__}.from_bytes())."
        )
        super().__init__(*args, **kwargs)
        self._dtype = np.uint8

        if ext is not None:
            if not ext.startswith("."):
                ext = "." + ext
            ext = ext.lower()
        self._ext = ext

        if size is not None:
            assert (
                len(size) == 2 and 0 < size[0] and 0 < size[1]
            ), f"Invalid image size info '{size}'"
            size = tuple(map(int, size))
        self._size = size  # (H, W)

    @classmethod
    def from_file(cls, path: str, *args, **kwargs):
        return ImageFromFile(path, *args, **kwargs)

    @classmethod
    def from_numpy(
        cls,
        data: Union[np.ndarray, Callable[[], np.ndarray]],
        *args,
        **kwargs,
    ):
        return ImageFromNumpy(data, *args, **kwargs)

    @classmethod
    def from_bytes(
        cls,
        data: Union[bytes, Callable[[], bytes]],
        *args,
        **kwargs,
    ):
        return ImageFromBytes(data, *args, **kwargs)

    @property
    def has_size(self) -> bool:
        """Indicates that size info is cached and won't require image loading"""
        return self._size is not None

    @property
    def size(self) -> Optional[Tuple[int, int]]:
        """Returns (H, W)"""

        if self._size is None:
            try:
                data = self.data
            except _image_loading_errors:
                return None
            if data is not None:
                self._size = tuple(map(int, data.shape[:2]))
        return self._size

    @property
    def ext(self) -> Optional[str]:
        """Media file extension (with the leading dot)"""
        return self._ext

    def _get_ext_to_save(self, fp: Union[str, io.IOBase], ext: Optional[str] = None):
        if isinstance(fp, str):
            assert ext is None, "'ext' must be empty if string is given."
            ext = osp.splitext(osp.basename(fp))[1].lower()
        else:
            ext = ext if ext else self._DEFAULT_EXT
        return ext

    def __eq__(self, other):
        # Do not compare `_type`
        # sicne Image is subclass of RoIImage and MosaicImage
        if not isinstance(other, __class__):
            return False
        return (np.array_equal(self.size, other.size)) and (np.array_equal(self.data, other.data))

    def set_crypter(self, crypter: Crypter):
        super().set_crypter(crypter)


class ImageFromFile(FromFileMixin, Image):
    def __init__(
        self,
        path: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(path, *args, **kwargs)
        self.__data = lazy_image(self.path, crypter=self._crypter)

        # extension from file name and real extension can be differ
        self._ext = self._ext if self._ext else osp.splitext(osp.basename(path))[1]

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""

        if not self.has_data:
            return None

        if self.__data._dtype != self._dtype:
            self.__data._loader = partial(load_image, dtype=self._dtype)
        data = self.__data()

        if self._size is None and data is not None:
            if not 2 <= data.ndim <= 3:
                raise MediaShapeError("An image should have 2 (gray) or 3 (bgra) dims.")
            self._size = tuple(map(int, data.shape[:2]))
        return data

    @property
    def size(self) -> Optional[Tuple[int, int]]:
        """Returns (H, W)"""

        if self._size is None:
            try:
                width, height = imagesize.get(self.path)
                assert width != -1 and height != -1
                self._size = (height, width)
            except Exception:
                _ = super().size
        return self._size

    def save(
        self,
        fp: Union[str, io.IOBase],
        ext: Optional[str] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        cur_path = osp.abspath(self.path) if self.path else None
        cur_ext = self.ext
        new_ext = self._get_ext_to_save(fp, ext)
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)

        if cur_path is not None and osp.isfile(cur_path):
            if cur_ext == new_ext:
                copyto_image(src=cur_path, dst=fp, src_crypter=self._crypter, dst_crypter=crypter)
            else:
                save_image(fp, self.data, ext=new_ext, crypter=crypter)
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), cur_path)

    def set_crypter(self, crypter: Crypter):
        super().set_crypter(crypter)
        if isinstance(self.__data, lazy_image):
            self.__data._crypter = crypter

    def get_data_as_dtype(self, dtype: Optional[np.dtype] = np.uint8) -> Optional[np.ndarray]:
        """Get image data with a specific data type"""
        self._dtype = dtype
        return self.data


class ImageFromData(FromDataMixin, Image):
    def save(
        self,
        fp: Union[str, io.IOBase],
        ext: Optional[str] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        data = self.data
        if data is None:
            raise ValueError(f"{self.__class__.__name__} is empty.")
        new_ext = self._get_ext_to_save(fp, ext)
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)
        save_image(fp, data, ext=new_ext, crypter=crypter)


class ImageFromNumpy(ImageFromData):
    def __init__(
        self,
        data: Union[Callable[[], bytes], bytes],
        *args,
        **kwargs,
    ):
        super().__init__(data=data, *args, **kwargs)

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""

        data = super().data

        if isinstance(data, np.ndarray) and data.dtype != self._dtype:
            data = np.clip(data, 0.0, 255.0).astype(self._dtype)
        if self._size is None and data is not None:
            if not 2 <= data.ndim <= 3:
                raise MediaShapeError("An image should have 2 (gray) or 3 (bgra) dims.")
            self._size = tuple(map(int, data.shape[:2]))
        return data

    @property
    def has_size(self) -> bool:
        """Indicates that size info is cached and won't require image loading"""
        return self._size is not None or isinstance(self._data, np.ndarray)

    def get_data_as_dtype(self, dtype: Optional[np.dtype] = np.uint8) -> Optional[np.ndarray]:
        """Get image data with a specific data type"""
        self._dtype = dtype
        return self.data


class ImageFromBytes(ImageFromData):
    _FORMAT_MAGICS = (
        (b"\x89PNG\r\n\x1a\n", ".png"),
        (b"\xff\xd8\xff", ".jpg"),
        (b"BM", ".bmp"),
    )

    def __init__(
        self,
        data: Union[Callable[[], bytes], bytes],
        *args,
        **kwargs,
    ):
        super().__init__(data=data, *args, **kwargs)

        if self._ext is None and isinstance(data, bytes):
            self._ext = self._guess_ext(data)

    @classmethod
    def _guess_ext(cls, data: bytes) -> Optional[str]:
        return next(
            (ext for magic, ext in cls._FORMAT_MAGICS if data.startswith(magic)),
            None,
        )

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""

        data = super().data

        if isinstance(data, bytes):
            data = decode_image(data, dtype=self._dtype)
        if self._size is None and data is not None:
            if not 2 <= data.ndim <= 3:
                raise MediaShapeError("An image should have 2 (gray) or 3 (bgra) dims.")
            self._size = tuple(map(int, data.shape[:2]))
        return data

    def get_data_as_dtype(self, dtype: Optional[np.dtype] = np.uint8) -> Optional[np.ndarray]:
        """Get image data with a specific data type"""

        if dtype != np.uint8:
            raise ValueError("ImageFromBytes only support `dtype=np.uint8`.")
        self._dtype = dtype
        return self.data


class VideoFrame(ImageFromNumpy):
    _type = MediaType.VIDEO_FRAME

    _DEFAULT_EXT = None

    def __init__(self, video: Video, index: int):
        self._video = video
        self._index = index

        super().__init__(data=lambda: self._video.get_frame_data(self._index))

    def as_dict(self) -> Dict[str, Any]:
        attrs = super().as_dict()
        return {
            "video": attrs["video"],
            "index": attrs["index"],
        }

    @property
    def size(self) -> Tuple[int, int]:
        return self._video.frame_size

    @property
    def index(self) -> int:
        return self._index

    @property
    def video(self) -> Video:
        return self._video

    @property
    def path(self) -> str:
        return self._video.path

    def from_self(self, **kwargs):
        attrs = deepcopy(self.as_dict())
        if "path" in kwargs:
            attrs.update({"video": self.video.from_self(**kwargs)})
            kwargs.pop("path")
        attrs.update(kwargs)
        return self.__class__(**attrs)

    def __getstate__(self):
        # Return only the picklable parts of the state.
        state = self.__dict__.copy()
        del state["_data"]
        return state

    def __setstate__(self, state):
        # Restore the objects' state.
        self.__dict__.update(state)
        # Reinitialize unpichlable attributes
        self._data = lambda: self._video.get_frame_data(self._index)


class _VideoFrameIterator(Iterator[VideoFrame]):
    """
    Provides sequential access to the video frames.
    """

    _video: Video
    _iterator: Iterator[VideoFrame]
    _pos: int
    _current_frame_data: Optional[np.ndarray]

    def __init__(self, video: Video):
        self._video = video
        self._reset()

    def _reset(self):
        self._video._reset_reader()
        self._iterator = self._decode(self._video._get_reader())
        self._pos = -1
        self._current_frame_data = None

    def _decode(self, cap) -> Iterator[VideoFrame]:
        """
        Decodes video frames using opencv
        """

        self._pos = -1

        success, frame = cap.read()
        while success:
            self._pos += 1
            if self._video._includes_frame(self._pos):
                self._current_frame_data = frame.astype(float)
                yield self._make_frame(index=self._pos)

            success, frame = cap.read()

        if self._video._frame_count is None:
            self._video._frame_count = self._pos + 1
            if self._video._end_frame and self._video._end_frame >= self._video._frame_count:
                raise ValueError(
                    f"The end_frame value({self._video._end_frame}) of the video "
                    f"must be less than the frame count({self._video._frame_count})."
                )

    def _make_frame(self, index) -> VideoFrame:
        return VideoFrame(self._video, index=index)

    def __next__(self):
        return next(self._iterator)

    def __getitem__(self, idx: int) -> VideoFrame:
        if not self._video._includes_frame(idx):
            raise IndexError(f"Video doesn't contain frame #{idx}.")

        return self._navigate_to(idx)

    def get_frame_data(self, idx: int) -> np.ndarray:
        self._navigate_to(idx)
        return self._current_frame_data

    def _navigate_to(self, idx: int) -> VideoFrame:
        """
        Iterates over frames to the required position.
        """

        if idx < 0:
            raise IndexError()

        if idx < self._pos:
            self._reset()

        if self._pos < idx:
            try:
                while self._pos < idx:
                    v = self.__next__()
            except StopIteration as e:
                raise IndexError() from e
        else:
            v = self._make_frame(index=self._pos)

        return v


class Video(MediaElement, Iterable[VideoFrame]):
    _type = MediaType.VIDEO

    """
    Provides random access to the video frames.
    """

    def __init__(
        self,
        path: str,
        step: int = 1,
        start_frame: int = 0,
        end_frame: Optional[int] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._path = path

        assert 0 <= start_frame
        if end_frame is not None:
            assert start_frame <= end_frame
            # we can't know the video length here,
            # so we cannot validate if the end_frame is valid.
        assert 0 < step
        self._step = step
        self._start_frame = start_frame
        self._end_frame = end_frame

        self._reader = None
        self._iterator: Optional[_VideoFrameIterator] = None
        self._frame_size: Optional[Tuple[int, int]] = None

        # We don't provide frame count unless we have a reliable source of
        # this information.
        # - Not all videos provide length / duration metainfo
        # - We can get an estimation based on metadata, but it
        #   can be invalid or inaccurate due to variable frame rate
        #   or fractional values rounded up. Relying on the value will give
        #   errors during requesting frames.
        # https://stackoverflow.com/a/47796468
        self._frame_count = None
        self._length = None

    def close(self):
        self._iterator = None

        if self._reader is not None:
            self._reader.release()
            self._reader = None

    def __getitem__(self, idx: int) -> VideoFrame:
        if not self._includes_frame(idx):
            raise IndexError(f"Video doesn't contain frame #{idx}.")

        return self._get_iterator()[idx]

    def get_frame_data(self, idx: int) -> VideoFrame:
        if not self._includes_frame(idx):
            raise IndexError(f"Video doesn't contain frame #{idx}.")

        return self._get_iterator().get_frame_data(idx)

    def __iter__(self) -> Iterator[VideoFrame]:
        """
        Iterates over frames lazily, if possible.
        """

        if self._frame_count is not None:
            # Decoding is not necessary to get frame pointers
            # However, it can be inacurrate
            end_frame = self._get_end_frame()
            for index in range(self._start_frame, end_frame + 1, self._step):
                yield VideoFrame(video=self, index=index)
        else:
            # Need to decode to iterate over frames
            yield from self._get_iterator()

    @property
    def length(self) -> Optional[int]:
        """
        Returns frame count of the closed interval [start_frame, end_frame],
        if video provides such information.

        Note that not all videos provide length / duration metainfo, so the
        result may be undefined.

        Also note, that information may be inaccurate because of variable
        FPS in video or incorrect metainfo. The count is only guaranteed to
        be valid after video is completely read once.

        The count is affected by the frame filtering options of the object,
        i.e. start frame, end frame and frame step.
        """

        if self._length is None:
            end_frame = self._get_end_frame()

            if end_frame is not None:
                length = (end_frame + 1 - self._start_frame) // self._step
                if 0 >= length:
                    raise ValueError(
                        "There is no valid frame for the closed interval"
                        f"[start_frame({self._start_frame}),"
                        f" end_frame({end_frame})] with step({self._step})."
                    )
                self._length = length

        return self._length

    @property
    def frame_size(self) -> Tuple[int, int]:
        """Returns (H, W)"""

        if self._frame_size is None:
            self._frame_size = self._get_frame_size()
        return self._frame_size

    def _get_frame_size(self) -> Tuple[int, int]:
        cap = self._get_reader()
        w = cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        h = cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

        if h and w:
            frame_size = (int(h), int(w))
        else:
            image = next(self._get_iterator()).data
            frame_size = image.shape[0:2]

        return frame_size

    def _get_end_frame(self):
        # Note that end_frame could less than the last frame of the video
        if self._end_frame is not None and self._frame_count is not None:
            end_frame = min(self._end_frame, self._frame_count)
        elif self._end_frame is not None:
            end_frame = self._end_frame
        elif self._frame_count is not None:
            end_frame = self._frame_count - 1
        else:
            end_frame = None

        return end_frame

    def _includes_frame(self, i):
        if self._start_frame <= i:
            if (i - self._start_frame) % self._step == 0:
                end_frame = self._get_end_frame()
                if end_frame is None or i <= end_frame:
                    return True

        return False

    def _get_iterator(self):
        if self._iterator is None:
            self._iterator = _VideoFrameIterator(self)
        return self._iterator

    def _get_reader(self):
        if self._reader is None:
            self._reset_reader()
        return self._reader

    def _reset_reader(self):
        if self._reader is not None:
            self._reader.release()
        self._reader = cv2.VideoCapture(self._path)
        assert self._reader.isOpened()

    def __eq__(self, other: object) -> bool:
        def _get_frame(obj: Video, idx: int):
            try:
                return obj[idx]
            except IndexError:
                return None

        if not isinstance(other, __class__):
            return False
        if self._start_frame != other._start_frame or self._step != other._step:
            return False

        # The video path can vary if a dataset is copied.
        # So, we need to check if the video data is the same instead of checking paths.
        if self._end_frame is not None and self._end_frame == other._end_frame:
            for idx in range(self._start_frame, self._end_frame + 1, self._step):
                if self[idx] != other[idx]:
                    return False
            return True

        end_frame = self._end_frame or other._end_frame
        if end_frame is None:
            last_frame = None
            for idx, frame in enumerate(self):
                if frame != _get_frame(other, frame.index):
                    return False
                last_frame = frame
            # check if the actual last frames are same
            try:
                other[last_frame.index + self._step if last_frame else self._start_frame]
            except IndexError:
                return True
            return False

        # _end_frame values, only one of the two is valid
        for idx in range(self._start_frame, end_frame + 1, self._step):
            frame = _get_frame(self, idx)
            if frame is None:
                return False
            if frame != _get_frame(other, idx):
                return False
        # check if the actual last frames are same
        idx_next = end_frame + self._step
        return None is (_get_frame(self, idx_next) or _get_frame(other, idx_next))

    def __hash__(self):
        # Required for caching
        return hash((self._path, self._step, self._start_frame, self._end_frame))

    def save(
        self,
        fp: Union[str, io.IOBase],
        crypter: Crypter = NULL_CRYPTER,
    ):
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)
        if isinstance(fp, str):
            if fp != self.path:
                shutil.copyfile(self.path, fp)
        elif isinstance(fp, io.IOBase):
            with open(self.path, "rb") as f_video:
                fp.write(f_video.read())

    @property
    def path(self) -> str:
        """Path to the media file"""
        return self._path

    @property
    def ext(self) -> str:
        """Media file extension (with the leading dot)"""
        return osp.splitext(osp.basename(self.path))[1]


class PointCloud(MediaElement[bytes]):
    _type = MediaType.POINT_CLOUD

    def __init__(
        self,
        extra_images: Optional[Union[List[Image], Callable[[], List[Image]]]] = None,
        *args,
        **kwargs,
    ):
        assert self.__class__ != PointCloud, (
            f"Directly initalizing {self.__class__.__name__} is not supported. "
            f"Please use one of fractory function ({self.__class__.__name__}.from_file(), "
            f"{self.__class__.__name__}.from_bytes())."
        )
        super().__init__(*args, **kwargs)
        self._extra_images = extra_images or []

    @classmethod
    def from_file(cls, path: str, *args, **kwargs):
        return PointCloudFromFile(path, *args, **kwargs)

    @classmethod
    def from_bytes(cls, data: Union[bytes, Callable[[], bytes]], *args, **kwargs):
        return PointCloudFromBytes(data, *args, **kwargs)

    @property
    def extra_images(self) -> List[Image]:
        if callable(self._extra_images):
            extra_images = self._extra_images()
            assert isinstance(extra_images, list) and all(
                [isinstance(image, Image) for image in extra_images]
            )
            return extra_images
        return self._extra_images

    def _save_extra_images(
        self,
        fn: Callable[[int, Image], Dict[str, Any]],
        crypter: Optional[Crypter] = None,
    ):
        crypter = crypter if crypter else self._crypter
        for i, img in enumerate(self.extra_images):
            if img.has_data:
                kwargs: Dict[str, Any] = {"crypter": crypter}
                kwargs.update(fn(i, img))
                img.save(**kwargs)

    def __eq__(self, other: object) -> bool:
        return (
            super().__eq__(other)
            and (self.data == other.data)
            and self.extra_images == other.extra_images
        )


class PointCloudFromFile(FromFileMixin, PointCloud):
    @property
    def data(self) -> Optional[bytes]:
        if self.has_data:
            with open(self.path, "rb") as f:
                bytes_data = f.read()
            return bytes_data
        return None

    def save(
        self,
        fp: Union[str, io.IOBase],
        extra_images_fn: Optional[Callable[[int, Image], Dict[str, Any]]] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        if not crypter.is_null_crypter:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not implement save() with non NullCrypter."
            )

        cur_path = osp.abspath(self.path) if self.path else None

        if cur_path is not None and osp.isfile(cur_path):
            with open(cur_path, "rb") as reader:
                _bytes = reader.read()
            if isinstance(fp, str):
                os.makedirs(osp.dirname(fp), exist_ok=True)
                with open(fp, "wb") as f:
                    f.write(_bytes)
            else:
                fp.write(_bytes)
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), cur_path)

        if extra_images_fn is not None:
            self._save_extra_images(extra_images_fn, crypter)


class PointCloudFromData(FromDataMixin, PointCloud):
    def save(
        self,
        fp: Union[str, io.IOBase],
        extra_images_fn: Optional[Callable[[int, Image], Dict[str, Any]]] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        if not crypter.is_null_crypter:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not implement save() with non NullCrypter."
            )

        _bytes = self.data
        if _bytes is None:
            raise ValueError(f"{self.__class__.__name__} is empty.")
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)
            with open(fp, "wb") as f:
                f.write(_bytes)
        else:
            fp.write(_bytes)

        if extra_images_fn is not None:
            self._save_extra_images(extra_images_fn, crypter)


class PointCloudFromBytes(PointCloudFromData):
    @property
    def data(self) -> Optional[bytes]:
        return super().data


class MultiframeImage(MediaElement):
    _type = MediaType.MULTIFRAME_IMAGE

    def __init__(
        self,
        images: Optional[Iterable[Union[str, Image, np.ndarray, Callable[[str], np.ndarray]]]],
        *,
        path: Optional[str] = None,
    ):
        self._path = path

        if images is None:
            images = []

        self._images = [None] * len(images)
        for i, image in enumerate(images):
            assert isinstance(image, (str, Image, np.ndarray)) or callable(image)

            if isinstance(image, str):
                image = Image.from_file(path=image)
            elif isinstance(image, np.ndarray) or callable(image):
                image = Image.from_numpy(data=image)

            self._images[i] = image

        assert self._path or self._images

    @property
    def data(self) -> List[Image]:
        return self._images

    @property
    def path(self) -> str:
        """Path to the media file"""
        return self._path

    @property
    def ext(self) -> str:
        """Media file extension (with the leading dot)"""
        return osp.splitext(osp.basename(self.path))[1]


class RoIImage(Image):
    _type = MediaType.ROI_IMAGE

    def __init__(
        self,
        roi: BboxIntCoords,
        *args,
        **kwargs,
    ):
        assert self.__class__ != RoIImage, (
            f"Directly initalizing {self.__class__.__name__} is not supported. "
            f"Please use a fractory function '{self.__class__.__name__}.from_image()'. "
        )

        assert len(roi) == 4 and all(isinstance(v, int) for v in roi)
        self._roi = roi
        _, _, w, h = self._roi
        super().__init__(size=(h, w), *args, **kwargs)

    def as_dict(self) -> Dict[str, Any]:
        attrs = super().as_dict()
        attrs.pop("size", None)
        return attrs

    @classmethod
    def from_file(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image'.")

    @classmethod
    def from_image(cls, data: Image, roi: BboxIntCoords, *args, **kwargs):
        if not isinstance(data, Image):
            raise TypeError(f"type(image)={type(data)} should be Image.")

        if isinstance(data, ImageFromFile):
            return RoIImageFromFile(path=data.path, roi=roi, ext=data._ext, *args, **kwargs)
        if isinstance(data, ImageFromNumpy):
            return RoIImageFromNumpy(data=data._data, roi=roi, ext=data._ext, *args, **kwargs)
        if isinstance(data, ImageFromBytes):
            return RoIImageFromBytes(data=data._data, roi=roi, ext=data._ext, *args, **kwargs)
        raise NotImplementedError

    @classmethod
    def from_numpy(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image'.")

    @classmethod
    def from_bytes(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image'.")

    @property
    def roi(self) -> BboxIntCoords:
        return self._roi

    def _get_roi_data(self, data: np.ndarray) -> np.ndarray:
        x, y, w, h = self._roi
        return data[y : y + h, x : x + w]

    def save(
        self,
        fp: Union[str, io.IOBase],
        ext: Optional[str] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        if not crypter.is_null_crypter:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not implement save() with non NullCrypter."
            )
        data = self.data
        if data is None:
            raise ValueError(f"{self.__class__.__name__} is empty.")
        new_ext = self._get_ext_to_save(fp, ext)
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)
        save_image(fp, data, ext=new_ext, crypter=crypter)


class RoIImageFromFile(FromFileMixin, RoIImage):
    def __init__(
        self,
        path: str,
        roi: BboxIntCoords,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(path, roi, *args, **kwargs)
        self.__data = lazy_image(self.path, crypter=self._crypter)

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""
        if not self.has_data:
            return None
        data = self.__data()
        return self._get_roi_data(data)


class RoIImageFromData(FromDataMixin, RoIImage):
    pass


class RoIImageFromBytes(RoIImageFromData):
    def __init__(
        self,
        data: Union[bytes, Callable[[], bytes]],
        roi: BboxIntCoords,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(data, roi, *args, **kwargs)

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""
        data = super().data
        if data is None:
            return None
        if isinstance(data, bytes):
            data = decode_image(data)
        return self._get_roi_data(data)


class RoIImageFromNumpy(RoIImageFromData):
    def __init__(
        self,
        data: Union[np.ndarray, Callable[[], np.ndarray]],
        roi: BboxIntCoords,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(data, roi, *args, **kwargs)

    @property
    def data(self) -> Optional[np.ndarray]:
        """Image data in BGRA HWC [0; 255] (uint8) format"""
        data = super().data
        if data is None:
            return None
        return self._get_roi_data(data)


ImageWithRoI = Tuple[Image, BboxIntCoords]


class MosaicImage(Image):
    _type = MediaType.MOSAIC_IMAGE

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        assert self.__class__ != MosaicImage, (
            f"Directly initalizing {self.__class__.__name__} is not supported. "
            f"Please use a fractory function '{self.__class__.__name__}.from_image_roi_pairs()'."
        )
        super().__init__(*args, **kwargs)

    @classmethod
    def from_file(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image_roi_pairs'.")

    @classmethod
    def from_image_roi_pairs(cls, data: List[ImageWithRoI], size: Tuple[int, int], *args, **kwargs):
        return MosaicImageFromImageRoIPairs(data, size)

    @classmethod
    def from_numpy(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image_roi_pairs'.")

    @classmethod
    def from_bytes(cls, *args, **kwargs):
        raise DatumaroError(f"Please use a factory function '{cls.__name__}.from_image_roi_pairs'.")


class MosaicImageFromData(FromDataMixin, MosaicImage):
    def save(
        self,
        fp: Union[str, io.IOBase],
        ext: Optional[str] = None,
        crypter: Crypter = NULL_CRYPTER,
    ):
        if not crypter.is_null_crypter:
            raise NotImplementedError(
                f"{self.__class__.__name__} does not implement save() with non NullCrypter."
            )
        data = self.data
        if data is None:
            raise ValueError(f"{self.__class__.__name__} is empty.")
        new_ext = self._get_ext_to_save(fp, ext)
        if isinstance(fp, str):
            os.makedirs(osp.dirname(fp), exist_ok=True)
        save_image(fp, data, ext=new_ext, crypter=crypter)


class MosaicImageFromImageRoIPairs(MosaicImageFromData):
    def __init__(self, data: List[ImageWithRoI], size: Tuple[int, int]) -> None:
        def _get_mosaic_img() -> np.ndarray:
            h, w = self.size
            mosaic_img = np.zeros(shape=(h, w, 3), dtype=np.uint8)
            for img, roi in data:
                assert isinstance(img, Image), "MosaicImage can only take a list of Images."
                x, y, w, h = roi
                mosaic_img[y : y + h, x : x + w] = img.data
            return mosaic_img

        super().__init__(data=_get_mosaic_img, size=size)
        self._data_in = data

    def as_dict(self) -> Dict[str, Any]:
        attrs = super().as_dict()
        return {
            "data": attrs["data_in"],
            "size": attrs["size"],
        }


TableDtype = TypeVar("TableDtype", str, int, float)


class Table:
    def __init__(
        self,
    ) -> None:
        """
        Table data with multiple rows and columns.
        This provides random access to the table row.

        Initialization must be done in the child class.
        """
        assert self.__class__ != Table, (
            f"Directly initalizing {self.__class__.__name__} is not supported. "
            f"Please use one of fractory functions ({self.__class__.__name__}.from_csv(), "
            f"{self.__class__.__name__}.from_dataframe(), or ({self.__class__.__name__}.from_list())."
        )
        self._shape: Tuple[int, int] = (0, 0)

    @classmethod
    def from_csv(cls, path: str, *args, **kwargs) -> Type[Table]:
        """
        Returns Table instance creating from a csv file.

        Args:
            path (str) : Path to csv file.
        """
        return TableFromCSV(path, *args, **kwargs)

    @classmethod
    def from_dataframe(
        cls,
        data: Union[pd.DataFrame, Callable[[], pd.DataFrame]],
        *args,
        **kwargs,
    ) -> Type[Table]:
        """
        Returns Table instance creating from a pandas DataFrame.

        Args:
            data (DataFrame) : Data in pandas DataFrame format.
        """
        return TableFromDataFrame(data, *args, **kwargs)

    @classmethod
    def from_list(
        cls,
        data: List[Dict[str, TableDtype]],
        *args,
        **kwargs,
    ) -> Type[Table]:
        """
        Returns Table instance creating from a list of dicts.

        Args:
            data (list(dict(str,str|int|float))) : A list of table row data.
        """
        return TableFromListOfDict(data, *args, **kwargs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.data.equals(other)

    def __getitem__(self, idx: int) -> TableRow:
        """
        Random access to a specific row by index.
        """
        if idx >= self.shape[0]:
            raise IndexError(f"Table doesn't contain row #{idx}.")
        return TableRow(table=self, index=idx)

    def __iter__(self) -> Iterator[TableRow]:
        """
        Iterates over rows.
        """
        for index in range(self.shape[0]):
            yield TableRow(table=self, index=index)

    @property
    def shape(self) -> Tuple[int, int]:
        """Returns table size as (#rows, #cols)"""
        return self._shape

    @property
    def columns(self) -> List[str]:
        """Returns column names"""
        return self.data.columns.to_list()

    def dtype(self, column: str) -> Optional[Type[TableDtype]]:
        """Returns native python type for a given column"""
        numpy_type = self.data.dtypes[column]
        if self.data[column].nunique() / self.shape[0] < 0.1:  # TODO
            # Convert to CategoricalDtype for efficient storage and categorical analysis
            return pd.api.types.CategoricalDtype()
        if isinstance(numpy_type, np.dtypes.ObjectDType):
            return str
        else:
            return type(np.zeros(1, numpy_type).tolist()[0])

    def features(self, column: str, unique: Optional[bool] = False) -> List[TableDtype]:
        """Get features for a given column name."""
        if unique:
            return list(self.data[column].unique())
        else:
            return self.data[column].to_list()

    def save(
        self,
        path: str,
    ):
        """
        Save table instance to a '.csv' file.

        Args:
            path (str) : Path to the output csv file.
        """
        data: pd.DataFrame = self.data
        os.makedirs(osp.dirname(path), exist_ok=True)
        data.to_csv(path, index=False)


class TableFromCSV(FromFileMixin, Table):
    def __init__(
        self,
        path: str,
        dtype: Optional[Dict] = None,
        sep: Optional[str] = None,
        encoding: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        """
        Read a '.csv' file and compose a Table instance.

        Args:
            path (str) : Path to csv file.
            dtype (optional, dict(str,str)) : Dictionay of column name -> type str ('str', 'int', or 'float').
            sep (optional, str) : Delimiter to use.
            encoding (optional, str) : Encoding to use for UTF when reading/writing (ex. 'utf-8').
        """
        super().__init__(path, *args, **kwargs)

        # assumes that the 1st row is a header.
        data: pd.DataFrame = pd.read_csv(
            path, dtype=dtype, sep=sep, engine="python", encoding=encoding, index_col=False
        )
        if data is None:
            raise ValueError(f"Can't read csv File from {path}")
        if data.shape[1] == 0:
            raise MediaShapeError("A table should have 1 or more columns.")

        self.__data = data
        self._shape = data.shape

    @property
    def data(self) -> Optional[pd.DataFrame]:
        """Table data in pandas DataFrame format"""
        return self.__data

    def select(self, columns: List[str]):
        self.__data = self.__data[columns]
        self._shape = self.__data.shape


class TableFromDataFrame(FromDataMixin, Table):
    def __init__(
        self,
        data: Union[Callable[[], pd.DataFrame], pd.DataFrame],
        *args,
        **kwargs,
    ):
        """
        Read a pandas DataFrame and compose a Table instance.

        Args:
            data (DataFrame) : Data in pandas DataFrame format.
        """
        super().__init__(data=data, *args, **kwargs)

        if data is None:
            raise ValueError("'data' can't be None")
        if data.shape[1] == 0:
            raise MediaShapeError("A table should have 1 or more columns.")
        for col in data.columns:
            if not isinstance(col, str):
                raise TypeError("A table should have column names as a list of str values")

        self._shape = data.shape

    @property
    def data(self) -> Optional[pd.DataFrame]:
        """Table data in pandas DataFrame format"""
        return super().data


class TableFromListOfDict(TableFromDataFrame):
    def __init__(
        self,
        data: List[Dict[str, TableDtype]],
        *args,
        **kwargs,
    ):
        """
        Read a list of table row data and compose a Table instance.
        The table row data is in dictionary format.

        Args:
            data (list(dict(str,str|int|float))) : A list of table row data.
        """
        super().__init__(data=pd.DataFrame(data), *args, **kwargs)


class TableRow(MediaElement):
    _type = MediaType.TABLE_ROW

    def __init__(self, table: Table, index: int):
        """
        TableRow media refers to a Table instance and its row index.

        Args:
            table (Table) : Table instance.
            index (int) : Row index.
        """
        if table is None:
            raise ValueError("'table' can't be None")
        if index < 0 or index >= table.shape[0]:
            raise IndexError(f"'index({index})' is out of range.")
        self._table = table
        self._index = index

    @property
    def table(self) -> Table:
        """Table instance"""
        return self._table

    @property
    def index(self) -> int:
        """Row index"""
        return self._index

    @property
    def path(self) -> str:
        return self._table.data.path

    @property
    def has_data(self) -> bool:
        return self.data() is not None

    def data(self, targets: Optional[List[str]] = None) -> Dict:
        """
        Row data in dict format.

        Args:
            targets (optional, list(str)) : If this is specified,
                the values corresponding to target colums will be returned.
                Otherwise, whole row data will be returned.
        """
        row = self.table.data.iloc[self.index]
        if targets:
            row = row[targets]
        return row.to_dict()

    def __repr__(self):
        return f"TableRow(row_idx:{self.index}, data:{self.data()})"

    @classmethod
    def from_data(cls, data: Dict, *args, **kwargs):
        return TableRowFromData(data, *args, **kwargs)


class TableRowFromData(FromDataMixin, TableRow):
    def __init__(self, data: Dict, *args, **kwargs):
        super().__init__(data=data, *args, **kwargs)

    @property
    def data(self) -> Dict:
        data = super().data
        return data



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.errors import ProjectNotFoundError
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Updates dataset from another one",
        description="""
        Updates items of the first dataset with items from the second one.|n
        |n
        By default, datasets are updated in-place. The '-o/--output-dir'
        option can be used to specify another output directory. When
        updating in-place, use the '--overwrite' parameter along with the
        '--save-media' export option (in-place updates fail by default
        to prevent data loss).|n
        |n
        Unlike the regular project data source joining, the datasets are not
        required to have the same labels. The labels from the "patch"
        dataset are projected onto the labels of the patched dataset,
        so only the annotations with the matching labels are used, i.e.
        all the annotations having unknown labels are ignored. Currently,
        this command doesn't allow to update the label information in the
        patched dataset.|n
        |n
        The command supports passing extra exporting options for the output
        dataset. The extra options should be passed after the main arguments
        and after the '--' separator. Particularly, this is useful to include
        images in the output dataset with '--save-media'.|n
        |n
        This command can be applied to the current project targets or
        arbitrary datasets outside a project. Note that if the target dataset
        is read-only (e.g. if it is a project, stage or a cache entry),
        the output directory must be provided.|n
        |n
        This command has the following invocation syntax:
        - %(prog)s <target dataset revpath> <patch dataset revpath>|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        The current project (-p/--project) is also used as a context for
        plugins, so it can be useful for dataset paths having custom formats.
        When not specified, the current project's working tree is used.|n
        |n
        Examples:|n
        - Update a VOC-like dataset with COCO-like annotations:|n
        |s|s%(prog)s --overwrite dataset1/:voc dataset2/:coco -- --save-media|n
        |n
        - Generate a patched dataset, based on a project:|n
        |s|s%(prog)s -o patched_proj1/ proj1/ proj2/|n
        |n
        - Update the "source1" source in the current project with a dataset:|n
        |s|s%(prog)s -p proj/ --overwrite source1 path/to/dataset2:coco|n
        |n
        - Generate a patched source from a previous revision and a dataset:|n
        |s|s%(prog)s -o new_src2/ HEAD~2:source-2 path/to/dataset2:yolo|n
        |n
        - Update a dataset in a custom format, described in a project plugin:|n
        |s|s%(prog)s -p proj/ --overwrite dataset/:my_format dataset2/:coco
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("target", help="Target dataset revpath")
    parser.add_argument("patch", help="Patch dataset revpath")
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        default=None,
        help="Output directory (default: save in-place)",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing files in the save directory, " "if it is not empty",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the 'current' project (default: current dir)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for exporting (pass '-- -h' for help). "
        "Must be specified after the main command arguments and after "
        "the '--' separator",
    )
    parser.set_defaults(command=patch_command)

    return parser


def get_sensitive_args():
    return {
        patch_command: ["target", "patch", "dst_dir", "project_dir", "extra_args"],
    }


@scoped
def patch_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    if project is not None:
        env = project.env
    else:
        env = DEFAULT_ENVIRONMENT

    target_dataset, _project = parse_full_revpath(args.target, project)
    if _project is not None:
        scope_add(_project)

    try:
        exporter = env.exporters[target_dataset.format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % args.format)

    extra_args = exporter.parse_cmdline(args.extra_args)

    dst_dir = args.dst_dir or target_dataset.data_path
    if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
        raise CliException(
            "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
        )
    dst_dir = osp.abspath(dst_dir)

    patch_dataset, _project = parse_full_revpath(args.patch, project)
    if _project is not None:
        scope_add(_project)

    target_dataset.update(patch_dataset)

    target_dataset.save(save_dir=dst_dir, **extra_args)

    log.info("Patched dataset has been saved to '%s'" % dst_dir)

    return 0



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# ruff: noqa: F405

from .dataset_item import *
from .media import *

__all__ = [
    # dataset_item
    "DatasetItemMapper",
    # media
    "MediaMapper",
]



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os.path as osp
import sys
import warnings

from ..util.telemetry_utils import (
    close_telemetry_session,
    init_telemetry_session,
    send_command_exception_info,
    send_command_failure_info,
    send_command_success_info,
)
from ..version import __version__
from . import contexts, helpers
from .commands import get_non_project_commands, get_project_commands
from .util import add_subparser, make_subcommands_help
from .util.errors import CliException

_log_levels = {
    "debug": log.DEBUG,
    "info": log.INFO,
    "warning": log.WARNING,
    "error": log.ERROR,
    "critical": log.CRITICAL,
}


def loglevel(name):
    return _log_levels[name]


class _LogManager:
    @classmethod
    def init_logger(cls, args=None):
        # Define minimalistic parser only to obtain loglevel
        parser = argparse.ArgumentParser(add_help=False)
        cls._define_loglevel_option(parser)
        args, _ = parser.parse_known_args(args)
        log_format = "%(asctime)s %(levelname)s: %(message)s"

        # Try setting up logging with basicConfig.
        # This does nothing, if other parts of the software
        # already configured handlers, i.e. during imports and when
        # main is called programmatically.
        log.basicConfig(format=log_format, level=args.loglevel)
        # Force-overwrite the log level and formatter
        log.root.setLevel(args.loglevel)
        for h in log.root.handlers:
            h.setFormatter(log.Formatter(log_format))

        # Suppress own deprecation warnings
        warnings.filterwarnings("ignore", category=DeprecationWarning, module=r"datumaro\..*")
        # We don't use sklearn directly, but it yells out too much about its deprecations.
        warnings.filterwarnings("ignore", category=DeprecationWarning, module=r"sklearn\..*")

    @staticmethod
    def _define_loglevel_option(parser):
        parser.add_argument(
            "--loglevel",
            type=loglevel,
            default="info",
            help="Logging level (options: %s; default: %s)"
            % (", ".join(_log_levels.keys()), "%(default)s"),
        )
        return parser


# TODO: revisit during CLI refactoring
def _get_known_contexts():
    return [
        ("model", contexts.model, "Actions with models"),
        ("project", contexts.project, "Actions with projects"),
        ("source", contexts.source, "Actions with data sources"),
        ("util", contexts.util, "Auxillary tools and utilities"),
    ]


def _get_helper_commands():
    return [
        (
            "format",
            helpers.format,
            "List and show information about supported import/export data formats",
        ),
    ]


def _get_sensitive_args():
    known_contexts = _get_known_contexts()
    known_commands = get_project_commands() + get_non_project_commands() + _get_helper_commands()

    res = {}
    for _, command, _ in known_contexts + known_commands:
        if command is not None:
            res.update(command.get_sensitive_args())

    return res


def make_parser():
    parser = argparse.ArgumentParser(
        description="Dataset Framework", formatter_class=argparse.RawDescriptionHelpFormatter
    )
    if parser.prog == osp.basename(__file__):  # python -m datumaro ...
        parser.prog = "datumaro"

    parser.add_argument("--version", action="version", version=__version__)
    _LogManager._define_loglevel_option(parser)

    known_contexts = _get_known_contexts()
    known_commands = get_non_project_commands()
    known_helpers = _get_helper_commands()

    # Argparse doesn't support subparser groups:
    # https://stackoverflow.com/questions/32017020/grouping-argparse-subparser-arguments
    help_line_start = max((len(e[0]) for e in known_contexts + known_commands), default=0)
    help_line_start = max((2 + help_line_start) // 4 + 1, 6) * 4  # align to tabs
    subcommands_desc = ""
    if known_contexts:
        subcommands_desc += "Contexts:\n"
        subcommands_desc += make_subcommands_help(known_contexts, help_line_start)
    if known_commands:
        if subcommands_desc:
            subcommands_desc += "\n"
        subcommands_desc += "Context-free Commands:\n"
        subcommands_desc += make_subcommands_help(known_commands, help_line_start)
    if known_helpers:
        if subcommands_desc:
            subcommands_desc += "\n"
        subcommands_desc += "Helper Commands:\n"
        subcommands_desc += make_subcommands_help(known_helpers, help_line_start)
    if subcommands_desc:
        subcommands_desc += (
            "\nRun '%s COMMAND --help' for more information on a command." % parser.prog
        )

    subcommands = parser.add_subparsers(
        title=subcommands_desc, description="", help=argparse.SUPPRESS
    )
    for command_name, command, _ in known_contexts + known_commands + known_helpers:
        if command is not None:
            add_subparser(subcommands, command_name, command.build_parser)

    return parser


def main(args=None):
    _LogManager.init_logger(args)

    parser = make_parser()
    args = parser.parse_args(args)

    if "command" not in args:
        parser.print_help()
        return 1

    sensitive_args = _get_sensitive_args()
    telemetry = init_telemetry_session(app_name="Datumaro", app_version=__version__)

    try:
        retcode = args.command(args)
        if retcode is None:
            retcode = 0
    except CliException as e:
        log.error(e)
        send_command_exception_info(telemetry, args, sensitive_args=sensitive_args[args.command])
        return 1
    except Exception as e:
        log.error(e)
        send_command_exception_info(telemetry, args, sensitive_args=sensitive_args[args.command])
        raise
    else:
        if retcode:
            send_command_failure_info(telemetry, args, sensitive_args=sensitive_args[args.command])
        else:
            send_command_success_info(telemetry, args, sensitive_args=sensitive_args[args.command])
        return retcode
    finally:
        close_telemetry_session(telemetry)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING, Sequence

from datumaro.components.dataset import Dataset

if TYPE_CHECKING:
    import datumaro.plugins.explorer as explorer
else:
    from datumaro.util.import_util import lazy_import

    explorer = lazy_import("datumaro.plugins.explorer")


class HashInference:
    def __init__(self, *datasets: Sequence[Dataset]) -> None:
        pass

    @property
    def model(self):
        if self._model is None:
            self._model = explorer.ExplorerLauncher(model_name="clip_visual_ViT-B_32")
        return self._model

    @property
    def text_model(self):
        if self._text_model is None:
            self._text_model = explorer.ExplorerLauncher(model_name="clip_text_ViT-B_32")
        return self._text_model

    def _compute_hash_key(self, datasets, datasets_to_infer):
        for dataset_to_infer in datasets_to_infer:
            if dataset_to_infer:
                dataset_to_infer.run_model(self.model, append_annotation=True)
        for dataset, dataset_to_infer in zip(datasets, datasets_to_infer):
            updated_items = [
                dataset.get(item.id, item.subset).wrap(annotations=item.annotations)
                for item in dataset_to_infer
            ]
            dataset.update(updated_items)
        return datasets



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ava_label.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0f\x61va_label.proto"\xa0\x01\n\x05Label\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08label_id\x18\x02 \x01(\r\x12$\n\nlabel_type\x18\x03 \x01(\x0e\x32\x10.Label.LabelType"Q\n\tLabelType\x12\x13\n\x0fPERSON_MOVEMENT\x10\x00\x12\x17\n\x13OBJECT_MANIPULATION\x10\x01\x12\x16\n\x12PERSON_INTERACTION\x10\x02""\n\tLabelList\x12\x15\n\x05label\x18\x01 \x03(\x0b\x32\x06.Labelb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "ava_label_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    _globals["_LABEL"]._serialized_start = 20
    _globals["_LABEL"]._serialized_end = 180
    _globals["_LABEL_LABELTYPE"]._serialized_start = 99
    _globals["_LABEL_LABELTYPE"]._serialized_end = 180
    _globals["_LABELLIST"]._serialized_start = 182
    _globals["_LABELLIST"]._serialized_end = 216
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import csv
import errno
import os
import os.path as osp
from typing import List, Optional

import google.protobuf.text_format as text_format

from datumaro.components.annotation import AnnotationType, Bbox, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.os_util import find_files

from . import ava_label_pb2


class AvaPath:
    IMAGE_DIR = "frames"
    IMAGE_EXT = ".jpg"
    ANNOTATION_DIR = "annotations"
    ANNOTATION_EXT = ".csv"
    ANNOTATION_PREFIX = "ava_"
    ANNOTATION_VERSION = "_v2.2"
    LABEL_LIST = ANNOTATION_PREFIX + "action_list" + ANNOTATION_VERSION + ".pbtxt"
    PROPOSAL_EXT = ".pkl"


class AvaBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find CSV file", path)
        self._path = path

        if not subset:
            subset = (
                osp.splitext(osp.basename(path))[0]
                .replace(AvaPath.ANNOTATION_PREFIX, "")
                .replace(AvaPath.ANNOTATION_VERSION, "")
            )
        super().__init__(subset=subset, ctx=ctx)

        if path.endswith(osp.join(AvaPath.ANNOTATION_DIR, osp.basename(path))):
            self._rootpath = path.rsplit(AvaPath.ANNOTATION_DIR, maxsplit=1)[0]
        else:
            raise FileNotFoundError(
                errno.ENOENT,
                f"Annotation path ({path}) should be under the directory which is named {AvaPath.ANNOTATION_DIR}. "
                "If not, Datumaro fails to find the root path for this dataset.",
            )

        if self._rootpath and osp.isdir(osp.join(self._rootpath, AvaPath.IMAGE_DIR)):
            self._images_dir = osp.join(self._rootpath, AvaPath.IMAGE_DIR)
        else:
            raise FileNotFoundError(
                errno.ENOENT,
                f"Root path ({self._rootpath}) should contain the directory which is named {AvaPath.IMAGE_DIR}. "
                "If not, Datumaro fails to find the image directory path.",
            )

        self._infos = self._load_infos(osp.dirname(path))

        category_path = osp.join(self._rootpath, AvaPath.ANNOTATION_DIR, AvaPath.LABEL_LIST)
        self._categories = self._load_categories(category_path)

        self._items = self._load_items(path)

    def _load_infos(self, path):
        infos = {}
        for file in os.listdir(path):
            if file.endswith(AvaPath.PROPOSAL_EXT):
                name = file.split(".")[0].split("_")[-1]
                infos[name + "_proposals"] = file

        return infos

    def _load_categories(self, category_path):
        if not osp.exists(category_path):
            raise FileNotFoundError(
                errno.ENOENT,
                f"Label lists cannot be found in ({category_path}). "
                "If not, Datumaro fails to import AVA action dataset.",
            )

        with open(category_path, "r") as f:
            pbtxt_data = f.read()

        label_list = ava_label_pb2.LabelList()
        text_format.Parse(pbtxt_data, label_list)

        categories = LabelCategories()

        # dummy class for id 0 for ava data
        if label_list.label[0].label_id != 0:
            categories.add("no action")
        for node in label_list.label:
            categories.add(node.name)

        return {AnnotationType.label: categories}

    def _load_items(self, ann_file):
        items = {}
        with open(ann_file, "r", encoding="utf-8") as f:
            csvreader = csv.reader(f)
            datas = list(csvreader)
            for data in datas:
                video_id = data[0]
                timestamp = data[1]

                item_id = video_id + "/" + timestamp
                image_path = osp.join(
                    self._images_dir, video_id, item_id.replace("/", "_") + AvaPath.IMAGE_EXT
                )

                item = items.get(item_id)
                if item is None:
                    item = DatasetItem(
                        id=item_id,
                        subset=self._subset,
                        media=Image.from_file(path=image_path),
                    )
                    items[item_id] = item

                if "excluded_timestamps" in self._subset:
                    continue

                bbox = list(map(float, data[2:6]))  # (x1, y1, x2, y2)
                label = int(data[6])
                entity_id = int(data[7])

                anns = item.annotations
                anns.append(
                    Bbox(
                        x=bbox[0],
                        y=bbox[1],
                        w=bbox[2] - bbox[0],
                        h=bbox[3] - bbox[1],
                        label=label,
                        attributes={"track_id": entity_id},
                    )
                )
                self._ann_types.add(AnnotationType.bbox)

        return items.values()


class AvaImporter(Importer):
    _ANNO_EXT = "csv"

    @classmethod
    def find_sources(cls, path):
        ann_files = find_files(
            osp.join(path, AvaPath.ANNOTATION_DIR), exts=cls._ANNO_EXT, recursive=True, max_depth=1
        )

        sources = []
        for ann_file in ann_files:
            if AvaPath.ANNOTATION_PREFIX in ann_file:
                sources.append({"url": ann_file, "format": AvaBase.NAME})

        return sources

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        super().detect(context)
        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [f".{cls._ANNO_EXT}"]


class AvaExporter(Exporter):
    DEFAULT_IMAGE_EXT = AvaPath.IMAGE_EXT

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        save_dir = self._save_dir

        ann_dir = osp.join(save_dir, AvaPath.ANNOTATION_DIR)
        os.makedirs(ann_dir, exist_ok=True)

        frame_dir = osp.join(save_dir, AvaPath.IMAGE_DIR)

        if self._save_dataset_meta:
            label_categories = self._extractor.categories()[AnnotationType.label]
            message = ava_label_pb2.LabelList()
            for k, v in label_categories._indices.items():
                label = ava_label_pb2.Label(name=k, label_id=v)
                message.label.extend([label])

            # Since protobuf may not be possible to describe zero while it is valid,
            # the label with label_id=0 will be ignored in the written pbtxt.
            # But this is well interpreted as zero during reading the pbtxt.
            pbtxt_string = text_format.MessageToString(message)
            with open(osp.join(ann_dir, AvaPath.LABEL_LIST), "w") as f:
                f.write(pbtxt_string)

        for subset_name, subset in self._extractor.subsets().items():
            ann_file = osp.join(
                ann_dir,
                AvaPath.ANNOTATION_PREFIX
                + subset_name
                + AvaPath.ANNOTATION_VERSION
                + AvaPath.ANNOTATION_EXT,
            )
            with open(ann_file, mode="w", newline="", encoding="utf-8") as csvfile:
                csvwriter = csv.writer(csvfile)
                for item in subset:
                    item_row = item.id.split("/")

                    if self._save_media:
                        image_path = osp.join(
                            osp.join(frame_dir, item_row[0]),
                            item.id.replace("/", "_") + AvaPath.IMAGE_EXT,
                        )
                        self._save_image(
                            item,
                            path=image_path,
                        )

                    bboxes = [a for a in item.annotations if a.type == AnnotationType.bbox]
                    if bboxes:
                        for bbox in bboxes:
                            csvwriter.writerow(
                                item_row
                                + [
                                    bbox.x,
                                    bbox.y,
                                    bbox.x + bbox.w,
                                    bbox.y + bbox.h,
                                    bbox.label,
                                    bbox.attributes.get("track_id", 0),
                                ]
                            )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from typing import Optional

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.plugins.data_formats.datumaro.importer import DatumaroImporter

from .format import DatumaroBinaryPath


class DatumaroBinaryImporter(DatumaroImporter):
    PATH_CLS = DatumaroBinaryPath

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--encryption-key",
            type=str,
            default=None,
            help="If the dataset is encrypted, "
            "it (secret key) is needed to import the dataset. "
            "If the incorrect key is given, it cannot be imported."
            "Ignore this argument if your dataset does not require encryption.",
        )
        parser.add_argument(
            "--num-workers",
            type=int,
            default=0,
            help="The number of multi-processing workers for import. "
            "If num_workers = 0, do not use multiprocessing (default: %(default)s).",
        )
        return parser

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        annot_files = context.require_files(
            osp.join(DatumaroBinaryPath.ANNOTATIONS_DIR, "*" + DatumaroBinaryPath.ANNOTATION_EXT)
        )

        for annot_file in annot_files:
            with context.probe_text_file(
                annot_file,
                f"{annot_file} has no Datumaro binary format signature",
                is_binary_file=True,
            ) as f:
                signature = f.read(DatumaroBinaryPath.SIGNATURE_LEN)
                signature = signature.decode()
                DatumaroBinaryPath.check_signature(signature)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT


import struct

import pyarrow as pa

from datumaro.components.dataset_base import IDataset
from datumaro.errors import DatasetImportError
from datumaro.plugins.data_formats.datumaro.exporter import JsonWriter
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper


class DatumaroArrow:
    SIGNATURE = "signature:datumaro_arrow"
    VERSION = "2.0"

    MP_TIMEOUT = 300.0  # 5 minutes

    ID_FIELD = "id"
    SUBSET_FIELD = "subset"
    MEDIA_FIELD = "media"

    IMAGE_FIELD = pa.struct(
        [
            pa.field("has_bytes", pa.bool_()),
            pa.field("bytes", pa.binary()),
            pa.field("path", pa.string()),
            pa.field("size", pa.list_(pa.uint16(), 2)),
        ]
    )
    POINT_CLOUD_FIELD = pa.struct(
        [
            pa.field("has_bytes", pa.bool_()),
            pa.field("bytes", pa.binary()),
            pa.field("path", pa.string()),
            pa.field("extra_images", pa.list_(pa.field("image", IMAGE_FIELD))),
        ]
    )
    MEDIA_FIELD = pa.struct(
        [
            pa.field("type", pa.uint8()),
            pa.field("image", IMAGE_FIELD),
            pa.field("point_cloud", POINT_CLOUD_FIELD),
        ]
    )
    SCHEMA = pa.schema(
        [
            pa.field(ID_FIELD, pa.string()),
            pa.field(SUBSET_FIELD, pa.string()),
            pa.field("media", MEDIA_FIELD),
            pa.field("annotations", pa.binary()),
            pa.field("attributes", pa.binary()),
        ]
    )

    @classmethod
    def check_signature(cls, signature: str):
        if signature != cls.SIGNATURE:
            raise DatasetImportError(
                f"Input signature={signature} is not aligned with the ground truth signature={cls.SIGNATURE}"
            )

    @classmethod
    def check_schema(cls, schema: pa.lib.Schema):
        if not cls.SCHEMA.equals(schema):
            raise DatasetImportError(
                f"input schema is not aligned with the ground truth schema.\n"
                f"input schema: \n{schema.to_string()}\n"
                f"ground truth schema: \n{cls.SCHEMA.to_string()}\n"
            )

    @classmethod
    def check_version(cls, version: str):
        if version != cls.VERSION:
            raise DatasetImportError(
                f"Input version={version} is not aligned with the current data format version={cls.VERSION}"
            )

    @classmethod
    def create_schema_with_metadata(cls, extractor: IDataset):
        media_type = extractor.media_type()._type
        categories = JsonWriter.write_categories(extractor.categories())

        return cls.SCHEMA.with_metadata(
            {
                "signature": cls.SIGNATURE,
                "version": cls.VERSION,
                "infos": DictMapper.forward(extractor.infos()),
                "categories": DictMapper.forward(categories),
                "media_type": struct.pack("<I", int(media_type)),
            }
        )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import os.path as osp
import warnings
from collections import defaultdict
from glob import glob
from io import TextIOWrapper
from typing import Any, Dict, List, Type

from defusedxml import ElementTree

from datumaro.components.errors import DatasetImportError
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.merge.extractor_merger import ExtractorMerger


class RoboflowCocoImporter(Importer):
    FORMAT = "roboflow_coco"
    ANN_FILE_NAME = "_annotations.coco.json"

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> FormatDetectionConfidence:
        context.require_file(osp.join("*", cls.ANN_FILE_NAME))
        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def find_sources(cls, path):
        subset_paths = glob(osp.join(path, "**", cls.ANN_FILE_NAME), recursive=True)

        sources = []
        for subset_path in subset_paths:
            subset_name = osp.basename(osp.dirname(subset_path))
            sources.append(
                {"url": subset_path, "format": cls.FORMAT, "options": {"subset": subset_name}}
            )

        return sources

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [osp.splitext(cls.ANN_FILE_NAME)[1]]

    @property
    def can_stream(self) -> bool:
        return True

    def get_extractor_merger(self) -> Type[ExtractorMerger]:
        return ExtractorMerger


class RoboflowVocImporter(Importer):
    FORMAT = "roboflow_voc"
    FORMAT_EXT = ".xml"
    ANN_DIR_NAME = ""

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        with context.require_any():
            with context.alternative():
                cls._check_ann_file(
                    context.require_file("**/" + cls.ANN_DIR_NAME + "*" + cls.FORMAT_EXT), context
                )

        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def _check_ann_file(cls, fpath: str, context: FormatDetectionContext) -> None:
        with context.probe_text_file(
            fpath, "Requirements for the annotation file of voc format"
        ) as fp:
            cls._check_ann_file_impl(fp)

    @classmethod
    def _check_ann_file_impl(cls, fp: TextIOWrapper) -> bool:
        root = ElementTree.parse(fp).getroot()

        if root.tag != "annotation":
            raise DatasetImportError("Roboflow VOC format xml file should have the annotation tag.")

        if not root.find("source/database").text == "roboflow.ai":
            raise DatasetImportError(
                "Roboflow VOC format xml file should have the source/database with `roboflow.ai`."
            )

        return True

    @classmethod
    def _get_sources(cls, path: str) -> Dict[Any, List[Any]]:
        def _filter_ann_file(fpath: str):
            try:
                with open(fpath, "r") as fp:
                    return cls._check_ann_file_impl(fp)
            except DatasetImportError:
                return False

        sources = cls._find_sources_recursive(
            path,
            ext=cls.FORMAT_EXT,
            extractor_name="",
            dirname=cls.ANN_DIR_NAME,
            file_filter=_filter_ann_file,
            filename="**/*",
            max_depth=2,
            recursive=True,
        )
        if len(sources) == 0:
            return []

        return sources

    @classmethod
    def find_sources(cls, path: str) -> List[Dict[str, Any]]:
        sources = cls._get_sources(path)

        subsets = {}
        for source in sources:
            subset_name = osp.dirname(source["url"]).split(os.sep)[-1]
            subsets[subset_name] = osp.dirname(source["url"])

        sources = [
            {
                "url": url,
                "format": cls.FORMAT,
                "options": {
                    "subset": subset,
                },
            }
            for subset, url in subsets.items()
        ]

        return sources

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls.FORMAT_EXT]


class RoboflowYoloImporter(RoboflowVocImporter):
    FORMAT = "roboflow_yolo"
    FORMAT_EXT = ".txt"
    ANN_DIR_NAME = "labels/"

    @classmethod
    def _check_ann_file_impl(cls, fp: TextIOWrapper) -> bool:
        for line in fp:
            fields = line.rstrip("\n").split(" ")
            if len(fields) != 5:
                raise DatasetImportError(
                    f"Roboflow Yolo format txt file should have 5 fields for each line, "
                    f"but the read line has {len(fields)} fields: fields={fields}."
                )

            for field in fields:
                if not field.replace(".", "").isdigit():
                    raise DatasetImportError(f"Each field should be a number but fields={fields}.")

            # Check the first line only
            return True

        raise DatasetImportError("Empty file is not allowed.")

    @classmethod
    def find_sources(cls, path: str) -> List[Dict[str, Any]]:
        sources = cls._get_sources(path)

        subsets = defaultdict(list)
        for source in sources:
            subset_name = osp.dirname(source["url"]).split(os.sep)[-2]
            subsets[subset_name].append(source["url"])

        sources = [
            {
                "url": osp.dirname(osp.dirname(urls[0])),
                "format": cls.FORMAT,
                "options": {
                    "subset": subset,
                    "urls": urls,
                },
            }
            for subset, urls in subsets.items()
        ]

        return sources


class RoboflowYoloObbImporter(RoboflowYoloImporter):
    FORMAT = "roboflow_yolo_obb"
    FORMAT_EXT = ".txt"
    ANN_DIR_NAME = "labelTxt/"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        warnings.warn(
            f"FormatDetectionConfidence of '{cls.FORMAT}' is lowered because of 'dota' format support. "
            f"It will be deprecated in datumaro==1.8.0.",
            DeprecationWarning,
        )
        with context.require_any():
            with context.alternative():
                cls._check_ann_file(
                    context.require_file("**/" + cls.ANN_DIR_NAME + "*" + cls.FORMAT_EXT), context
                )

        return FormatDetectionConfidence.LOW

    @classmethod
    def _check_ann_file_impl(cls, fp: TextIOWrapper) -> bool:
        for line in fp:
            fields = line.rstrip("\n").split(" ")
            if len(fields) != 10:
                raise DatasetImportError(
                    f"Roboflow Yolo OBB format txt file should have 10 fields for each line, "
                    f"but the read line has {len(fields)} fields: fields={fields}."
                )

            # Check the first line only
            return True

        raise DatasetImportError("Empty file is not allowed.")


class RoboflowCreateMlImporter(RoboflowCocoImporter):
    FORMAT = "roboflow_create_ml"
    ANN_FILE_NAME = "_annotations.createml.json"


class RoboflowMulticlassImporter(RoboflowCocoImporter):
    FORMAT = "roboflow_multiclass"
    ANN_FILE_NAME = "_classes.csv"



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from abc import ABC, abstractclassmethod
from importlib import import_module
from importlib.util import find_spec
from typing import Dict, List, Optional, Sequence, Type, Union

from datumaro.components.dataset_base import DatasetBase
from datumaro.components.errors import DatumaroError
from datumaro.components.exporter import Exporter
from datumaro.components.generator import DatasetGenerator
from datumaro.components.importer import Importer
from datumaro.components.launcher import Launcher
from datumaro.components.transformer import Transform
from datumaro.components.validator import Validator

# Transform should be in front of DatasetBase,
# since Transform inherits DatasetBase.
_PLUGIN_TYPES = Union[
    Transform,
    Exporter,
    DatasetGenerator,
    Importer,
    Launcher,
    Validator,
    DatasetBase,
]
PLUGIN_TYPES = Type[_PLUGIN_TYPES]
STR_TO_PLUGIN_TYPES = {
    t.__name__: t
    for t in [
        Transform,
        Exporter,
        DatasetGenerator,
        Importer,
        Launcher,
        Validator,
        DatasetBase,
    ]
}
_EXTRA_DEPS_ATTR_NAME = "__extra_deps__"


class LazyPlugin(ABC):
    NAME: str

    @abstractclassmethod
    def get_plugin_cls(cls) -> PLUGIN_TYPES:
        pass


def get_lazy_plugin(
    import_path: str,
    plugin_name: str,
    plugin_type: str,
    extra_deps: List[str] = [],
    metadata: Dict = {},
) -> Optional[LazyPlugin]:
    for extra_dep in extra_deps:
        spec = find_spec(extra_dep)
        if spec is None:
            log.debug(f"Cannot import extra dep={extra_dep} for plugin_name={plugin_name}.")
            return None

    plugin_type_cls = STR_TO_PLUGIN_TYPES[plugin_type]

    class LazyPluginImpl(LazyPlugin, plugin_type_cls):
        NAME = plugin_name
        METADATA = metadata

        @classmethod
        def get_plugin_cls(cls) -> PLUGIN_TYPES:
            splits = import_path.split(".")
            module_name = ".".join(splits[:-1])
            class_name = splits[-1]
            module = import_module(module_name)
            plugin_cls = getattr(module, class_name)
            return plugin_cls

    return LazyPluginImpl


def extra_deps(*extra_dep_names: Sequence[str]):
    """Decorator to assign extra deps for the plugin class.

    There exist some plugins that cannot be executable with the default installation setup.
    For example, `AcLauncher` plugin needs `tensorflow` and `openvino.tools` extra dependencies.
    In this case, you have to add this decorator to that plugin class definition as follows.

    @extra_deps("tensorflow", "openvino.tools")
    class AcLauncher(Launcher, CliPlugin):
    ...
    """

    def inner(plugin_cls: object):
        if hasattr(plugin_cls, _EXTRA_DEPS_ATTR_NAME):
            raise DatumaroError(f"It already has {_EXTRA_DEPS_ATTR_NAME} attribute!")

        setattr(plugin_cls, _EXTRA_DEPS_ATTR_NAME, list(extra_dep_names))
        return plugin_cls

    return inner


def get_extra_deps(plugin_cls: object) -> List[str]:
    """Get extra deps of the given plugin class. If not exists, return an empty list."""
    return getattr(plugin_cls, _EXTRA_DEPS_ATTR_NAME, [])



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from datumaro.components.errors import MediaTypeError
from datumaro.components.media import Image
from datumaro.plugins.openvino_plugin.launcher import OpenvinoLauncher


class ShiftLauncher(OpenvinoLauncher):
    def __init__(
        self,
        description=None,
        weights=None,
        interpreter=None,
        model_dir=None,
        model_name=None,
        output_layers=None,
        device=None,
    ):
        super().__init__(
            description,
            weights,
            interpreter,
            model_dir,
            model_name,
            output_layers,
            device,
        )

        self._device = device or "cpu"
        self._output_blobs = next(iter(self._net.outputs))

    def type_check(self, item):
        if not isinstance(item.media, Image):
            raise MediaTypeError(f"Media type should be Image, Current type={type(item.media)}")
        return True



# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import os
import os.path as osp
from typing import List, Optional

import numpy as np

from datumaro.components.annotation import AnnotationType, Label, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.format_detection import FormatDetectionConfidence
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import Image
from datumaro.util.meta_file_util import has_meta_file, parse_meta_file


class MnistCsvPath:
    IMAGE_SIZE = 28
    NONE_LABEL = -1


class MnistCsvBase(SubsetBase):
    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        if not osp.isfile(path):
            raise FileNotFoundError(errno.ENOENT, "Can't find annotations file", path)

        if not subset:
            file_name = osp.splitext(osp.basename(path))[0]
            subset = file_name.rsplit("_", maxsplit=1)[-1]

        super().__init__(subset=subset, ctx=ctx)
        self._dataset_dir = osp.dirname(path)

        self._categories = self._load_categories()

        self._items = list(self._load_items(path).values())

    def _load_categories(self):
        if has_meta_file(self._dataset_dir):
            return {
                AnnotationType.label: LabelCategories.from_iterable(
                    parse_meta_file(self._dataset_dir).keys()
                )
            }

        label_cat = LabelCategories()

        labels_file = osp.join(self._dataset_dir, "labels.txt")
        if osp.isfile(labels_file):
            with open(labels_file, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    label_cat.add(line)
        else:
            for i in range(10):
                label_cat.add(str(i))

        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}
        with open(path, "r", encoding="utf-8") as f:
            annotation_table = f.readlines()

        metafile = osp.join(self._dataset_dir, "meta_%s.csv" % self._subset)
        meta = []
        if osp.isfile(metafile):
            with open(metafile, "r", encoding="utf-8") as f:
                meta = f.readlines()

        for i, data in enumerate(annotation_table):
            data = data.split(",")
            item_anno = []
            try:
                label = int(data[0])
            except ValueError:
                continue
            if label != MnistCsvPath.NONE_LABEL:
                item_anno.append(Label(label))
                self._ann_types.add(AnnotationType.label)

            if 0 < len(meta):
                meta[i] = meta[i].strip().split(",")

            # support for single-channel image only
            image = None
            if 1 < len(data):
                if 0 < len(meta) and 1 < len(meta[i]):
                    image = np.array([int(pix) for pix in data[1:]], dtype="uint8").reshape(
                        int(meta[i][-2]), int(meta[i][-1])
                    )
                else:
                    image = np.array([int(pix) for pix in data[1:]], dtype="uint8").reshape(28, 28)

            if image is not None:
                image = Image.from_numpy(data=image)

            if 0 < len(meta) and len(meta[i]) in [1, 3]:
                i = meta[i][0]

            items[i] = DatasetItem(id=i, subset=self._subset, media=image, annotations=item_anno)

        return items


class MnistCsvImporter(Importer):
    DETECT_CONFIDENCE = FormatDetectionConfidence.MEDIUM
    _ANNO_EXT = ".csv"

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(
            path,
            cls._ANNO_EXT,
            "mnist_csv",
            file_filter=lambda p: osp.basename(p).find("mnist_") != -1,
        )

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]


class MnistCsvExporter(Exporter):
    DEFAULT_IMAGE_EXT = ".png"

    def _apply_impl(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        os.makedirs(self._save_dir, exist_ok=True)
        if self._save_dataset_meta:
            self._save_meta_file(self._save_dir)

        for subset_name, subset in self._extractor.subsets().items():
            data = []
            item_ids = {}
            image_sizes = {}
            for item in subset:
                anns = [a.label for a in item.annotations if a.type == AnnotationType.label]
                label = MnistCsvPath.NONE_LABEL
                if anns:
                    label = anns[0]

                if item.media and self._save_media:
                    image = item.media
                    if not image.has_data:
                        data.append([label, None])
                    else:
                        if (
                            image.data.shape[0] != MnistCsvPath.IMAGE_SIZE
                            or image.data.shape[1] != MnistCsvPath.IMAGE_SIZE
                        ):
                            image_sizes[len(data)] = [image.data.shape[0], image.data.shape[1]]
                        image = image.data.reshape(-1).astype(np.uint8).tolist()
                        image.insert(0, label)
                        data.append(image)
                else:
                    data.append([label])

                if item.id != str(len(data) - 1):
                    item_ids[len(data) - 1] = item.id

            anno_file = osp.join(self._save_dir, "mnist_%s.csv" % subset_name)
            self.save_in_csv(anno_file, data)

            # it is't in the original format,
            # this is for storng other names and sizes of images
            if len(item_ids) or len(image_sizes):
                meta = []
                if len(item_ids) and len(image_sizes):
                    # other names and sizes of images
                    size = [MnistCsvPath.IMAGE_SIZE, MnistCsvPath.IMAGE_SIZE]
                    for i in range(len(data)):
                        w, h = image_sizes.get(i, size)
                        meta.append([item_ids.get(i, i), w, h])

                elif len(item_ids):
                    # other names of images
                    for i in range(len(data)):
                        meta.append([item_ids.get(i, i)])

                elif len(image_sizes):
                    # other sizes of images
                    size = [MnistCsvPath.IMAGE_SIZE, MnistCsvPath.IMAGE_SIZE]
                    for i in range(len(data)):
                        meta.append(image_sizes.get(i, size))

                metafile = osp.join(self._save_dir, "meta_%s.csv" % subset_name)
                self.save_in_csv(metafile, meta)

        self.save_labels()

    def save_in_csv(self, path, data):
        with open(path, "w", encoding="utf-8") as f:
            for row in data:
                f.write(",".join([str(p) for p in row]) + "\n")

    def save_labels(self):
        labels_file = osp.join(self._save_dir, "labels.txt")
        with open(labels_file, "w", encoding="utf-8") as f:
            f.writelines(
                l.name + "\n"
                for l in self._extractor.categories().get(AnnotationType.label, LabelCategories())
            )



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

# ruff: noqa: E501

import os
import os.path as osp
import re
from distutils.util import strtobool

import setuptools
from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools_rust import Binding, RustExtension


def find_version(project_dir=None):
    if not project_dir:
        project_dir = osp.dirname(osp.abspath(__file__))

    file_path = osp.join(project_dir, "datumaro", "version.py")

    with open(file_path, "r", encoding="utf-8") as version_file:
        version_text = version_file.read()

    # PEP440:
    # https://www.python.org/dev/peps/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions
    pep_regex = r"([1-9]\d*!)?(0|[1-9]\d*)(\.(0|[1-9]\d*))*((a|b|rc)(0|[1-9]\d*))?(\.post(0|[1-9]\d*))?(\.dev(0|[1-9]\d*))?"
    version_regex = r"__version__\s*=\s*.(" + pep_regex + ")."
    match = re.match(version_regex, version_text)
    if not match:
        raise RuntimeError("Failed to find version string in '%s'" % file_path)

    version = version_text[match.start(1) : match.end(1)]
    return version


CORE_REQUIREMENTS_FILE = "requirements-core.txt"
DEFAULT_REQUIREMENTS_FILE = "requirements-default.txt"


def parse_requirements(filename=CORE_REQUIREMENTS_FILE):
    with open(filename, "r", encoding="utf-8") as fh:
        return fh.readlines()


CORE_REQUIREMENTS = parse_requirements(CORE_REQUIREMENTS_FILE)
if strtobool(os.getenv("DATUMARO_HEADLESS", "0").lower()):
    CORE_REQUIREMENTS.append("opencv-python-headless")
else:
    CORE_REQUIREMENTS.append("opencv-python")

DEFAULT_REQUIREMENTS = parse_requirements(DEFAULT_REQUIREMENTS_FILE)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

ext_modules = [
    Pybind11Extension(
        "datumaro._capi",
        ["src/datumaro/capi/pybind.cpp"],
        define_macros=[("VERSION_INFO", find_version("./src"))],
        extra_compile_args=["-O3"],
    ),
]

setuptools.setup(
    name="datumaro",
    version=find_version("./src"),
    author="Intel",
    author_email="emily.chun@intel.com",
    description="Dataset Management Framework (Datumaro)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/openvinotoolkit/datumaro",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src", include=["datumaro*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    install_requires=CORE_REQUIREMENTS,
    extras_require={
        "tf": ["tensorflow"],
        "tfds": ["tensorflow-datasets<4.9.3"],
        "torch": ["torch", "torchvision", "torchtext==0.16.0"],
        "default": DEFAULT_REQUIREMENTS,
    },
    ext_modules=ext_modules,
    entry_points={
        "console_scripts": [
            "datum=datumaro.cli.__main__:main",
        ],
    },
    cmdclass={"build_ext": build_ext},
    package_data={
        "datumaro.plugins.synthetic_data": ["background_colors.txt"],
        "datumaro.plugins.openvino_plugin.samples": ["coco.class", "imagenet.class"],
    },
    include_package_data=True,
    rust_extensions=[
        RustExtension(
            "datumaro.rust_api",
            path=osp.join("rust", "Cargo.toml"),
            debug=False,
            binding=Binding.PyO3,
        ),
    ],
    # rust extensions are not zip safe, just like C-extensions.
    zip_safe=False,
)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import glob
import importlib
import logging as log
import os.path as osp
from functools import partial
from inspect import getmodule, isclass
from typing import Callable, List, Optional, Sequence, Set

from datumaro.components.cli_plugin import plugin_types
from datumaro.components.format_detection import (
    DetectedFormat,
    FormatDetectionConfidence,
    RejectionReason,
    detect_dataset_format,
)
from datumaro.components.registry import (
    DatasetBaseRegistry,
    ExporterRegistry,
    GeneratorRegistry,
    ImporterRegistry,
    LauncherRegistry,
    PluginRegistry,
    TransformRegistry,
    ValidatorRegistry,
)
from datumaro.util.os_util import get_all_file_extensions, import_foreign_module, split_path


class Environment:
    _builtin_plugins = None

    def __init__(self, use_lazy_import: bool = True):
        self._extractors = DatasetBaseRegistry()
        self._importers = ImporterRegistry()
        self._launchers = LauncherRegistry()
        self._exporters = ExporterRegistry()
        self._generators = GeneratorRegistry()
        self._transforms = TransformRegistry()
        self._validators = ValidatorRegistry()
        self._builtins_initialized = False
        self._use_lazy_import = use_lazy_import

    def _get_plugin_registry(self, name):
        if not self._builtins_initialized:
            self._builtins_initialized = True
            self._register_builtin_plugins()
        return getattr(self, name)

    @property
    def extractors(self) -> DatasetBaseRegistry:
        return self._get_plugin_registry("_extractors")

    @property
    def importers(self) -> ImporterRegistry:
        return self._get_plugin_registry("_importers")

    @property
    def launchers(self) -> LauncherRegistry:
        return self._get_plugin_registry("_launchers")

    @property
    def exporters(self) -> ExporterRegistry:
        return self._get_plugin_registry("_exporters")

    @property
    def generators(self) -> GeneratorRegistry:
        return self._get_plugin_registry("_generators")

    @property
    def transforms(self) -> TransformRegistry:
        return self._get_plugin_registry("_transforms")

    @property
    def validators(self) -> ValidatorRegistry:
        return self._get_plugin_registry("_validators")

    @staticmethod
    def _find_plugins(plugins_dir):
        plugins = []

        for pattern in ("*.py", "*/*.py", "*/*/*.py"):
            for path in glob.glob(osp.join(glob.escape(plugins_dir), pattern)):
                if not osp.isfile(path):
                    continue

                path_rel = osp.relpath(path, plugins_dir)
                name_parts = split_path(osp.splitext(path_rel)[0])

                # a module with a dot in the name won't load correctly
                if any("." in part for part in name_parts):
                    log.warning(
                        "Python file '%s' in directory '%s' can't be imported "
                        "due to a dot in the name; skipping.",
                        path_rel,
                        plugins_dir,
                    )
                    continue
                plugins.append(".".join(name_parts))

        return plugins

    @classmethod
    def _get_plugin_exports(cls, module, types):
        exports = []
        if hasattr(module, "exports"):
            exports = module.exports
        else:
            for symbol in dir(module):
                if symbol.startswith("_"):
                    continue
                exports.append(getattr(module, symbol))

        exports = [
            s
            for s in exports
            if isclass(s)
            and issubclass(s, types)
            and s not in types
            and (
                getmodule(s)
                is None  # Custom plugin (in the Datumaro project) can be a single file and have no module
                or not getmodule(s).__package__.startswith("datumaro.components")
            )
        ]

        return exports

    @classmethod
    def _load_plugins(cls, module_names, *, importer, types=None):
        types = tuple(types or plugin_types())

        all_exports = []
        for module_name in module_names:
            try:
                module = importer(module_name)
                exports = cls._get_plugin_exports(module, types)
            except Exception as e:
                module_search_error = ModuleNotFoundError

                message = ["Failed to import module '%s': %s", module_name, e]
                if isinstance(e, module_search_error):
                    log.debug(*message)
                else:
                    log.warning(*message)
                continue

            log.debug(
                "Imported the following symbols from %s: %s"
                % (module_name, ", ".join(s.__name__ for s in exports))
            )
            all_exports.extend(exports)

        return all_exports

    @classmethod
    def _load_builtin_plugins(cls):
        """Load builtin plugins which will be imported lazily using plugin spec files"""
        if cls._builtin_plugins is None:
            from datumaro.plugins.specs import get_lazy_plugins

            cls._builtin_plugins = get_lazy_plugins()
        return cls._builtin_plugins

    @classmethod
    def _load_builtin_plugins_from_importlib(cls):
        """Load builtin plugins from importlib, not lazy import from plugin spec files"""
        if cls._builtin_plugins is None:
            import datumaro.plugins

            plugins_dir = osp.dirname(datumaro.plugins.__file__)
            module_names = [
                datumaro.plugins.__name__ + "." + name for name in cls._find_plugins(plugins_dir)
            ]
            cls._builtin_plugins = cls._load_plugins(module_names, importer=importlib.import_module)
        return cls._builtin_plugins

    def load_plugins(self, plugins_dir):
        module_names = self._find_plugins(plugins_dir)
        plugins = self._load_plugins(
            module_names, importer=partial(import_foreign_module, path=plugins_dir)
        )

        self.register_plugins(plugins)

    def _register_builtin_plugins(self):
        self.register_plugins(
            self._load_builtin_plugins()
            if self._use_lazy_import
            else self._load_builtin_plugins_from_importlib()
        )

    def register_plugins(self, plugins):
        self.extractors.batch_register(plugins)
        self.importers.batch_register(plugins)
        self.launchers.batch_register(plugins)
        self.exporters.batch_register(plugins)
        self.generators.batch_register(plugins)
        self.transforms.batch_register(plugins)
        self.validators.batch_register(plugins)

    def make_extractor(self, name, *args, **kwargs):
        return self.extractors.get(name)(*args, **kwargs)

    def make_importer(self, name, *args, **kwargs):
        return self.importers.get(name)(*args, **kwargs)

    def make_launcher(self, name, *args, **kwargs):
        return self.launchers.get(name)(*args, **kwargs)

    def make_exporter(self, name, *args, **kwargs):
        result = self.exporters.get(name)
        if isclass(result):
            result = result.convert
        return partial(result, *args, **kwargs)

    def make_transform(self, name, *args, **kwargs):
        return partial(self.transforms.get(name), *args, **kwargs)

    def is_format_known(self, name):
        return name in self.importers or name in self.extractors

    def detect_dataset(
        self,
        path: str,
        depth: int = 1,
        rejection_callback: Optional[Callable[[str, RejectionReason, str], None]] = None,
    ) -> List[str]:
        ignore_dirs = {"__MSOSX", "__MACOSX"}
        all_matched_formats: Set[DetectedFormat] = set()

        extensions = get_all_file_extensions(path, ignore_dirs) or [""]

        importers = {
            (name, importer.get_plugin_cls() if self._use_lazy_import else importer)
            for extension in extensions
            for name, importer in self.importers.extension_groups.get(extension, [])
        }
        for _ in range(depth + 1):
            detected_formats = detect_dataset_format(
                ((format_name, importer.detect) for format_name, importer in importers),
                path,
                rejection_callback=rejection_callback,
            )

            if detected_formats:
                all_matched_formats |= set(detected_formats)

            paths = glob.glob(osp.join(path, "*"))
            path = "" if len(paths) != 1 else paths[0]
            if not osp.isdir(path) or osp.basename(path) in ignore_dirs:
                break

        max_conf = (
            max(all_matched_formats).confidence
            if len(all_matched_formats) > 0
            else FormatDetectionConfidence.NONE
        )

        return sorted(
            [str(format) for format in all_matched_formats if format.confidence == max_conf]
        )

    def __reduce__(self):
        return (self.__class__, ())

    @classmethod
    def merge(cls, envs: Sequence["Environment"]) -> "Environment":
        if all([env == DEFAULT_ENVIRONMENT for env in envs]):
            return DEFAULT_ENVIRONMENT

        merged = Environment()

        def _register(registry: PluginRegistry):
            merged.register_plugins(list(registry._items.values()))

        for env in envs:
            _register(env.extractors)
            _register(env.importers)
            _register(env.launchers)
            _register(env.exporters)
            _register(env.generators)
            _register(env.transforms)
            _register(env.validators)

        return merged

    @classmethod
    def release_builtin_plugins(cls):
        cls._builtin_plugins = None


DEFAULT_ENVIRONMENT = Environment()



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os.path as osp

from datumaro.components.algorithms.hash_key_inference.prune import Prune
from datumaro.components.errors import ProjectNotFoundError
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Prune dataset and make a representative subset",
        description="""
        Apply data pruning to a dataset.|n
        The command can be useful if you have to extract representative subset.
        |n
        The current project (-p/--project) is used as a context for plugins
        and models. It is used when there is a dataset path in target.
        When not specified, the current project's working tree is used.|n
        |n
        By default, datasets are updated in-place. The '-o/--output-dir'
        option can be used to specify another output directory. When
        updating in-place, use the '--overwrite' parameter (in-place
        updates fail by default to prevent data loss), unless a project
        target is modified.|n
        |n
        The command can be applied to a dataset or a project build target,
        a stage or the combined 'project' target, in which case all the
        targets will be affected.|n
        |n
        Examples:|n
        - Prune dataset with selecting random and ratio 80%:|n
        |s|s%(prog)s -m random -r 0.8|n
        - Prune dataset with clustering in image hash and ratio 50%:|n
        |s|s%(prog)s -m query_clust -h img -r 0.5|
        - Prune dataset based on entropy with clustering in image hash and ratio 50%:|n
        |s|s%(prog)s -m entropy -h img -r 0.5|
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target",
        nargs="?",
        default="project",
        metavar="revpath",
        help="Target dataset revpath (default: project)",
    )
    parser.add_argument("-m", "--method", dest="method", help="Method to apply to the dataset")
    parser.add_argument(
        "-r", "--ratio", type=float, dest="ratio", help="How much to remain dataset after pruning"
    )
    parser.add_argument(
        "--hash-type",
        type=str,
        dest="hash_type",
        default="img",
        help="Hashtype to extract feature from data information between image and text(label)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )

    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="""
            Output directory. Can be omitted for main project targets
            (i.e. data sources and the 'project' target, but not
            intermediate stages) and dataset targets.
            If not specified, the results will be saved inplace.
            """,
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.set_defaults(command=prune_command)

    return parser


def get_sensitive_args():
    return {
        prune_command: [
            "target",
            "method",
            "ratio",
            "hash_type",
            "project_dir",
            "dst_dir",
        ]
    }


@scoped
def prune_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    targets = [args.target] if args.target else list(project.working_tree.sources)

    source_dataset = [parse_full_revpath(target, project)[0] for target in targets][0]

    prune = Prune(source_dataset, cluster_method=args.method, hash_type=args.hash_type)
    result = prune.get_pruned(args.ratio)

    dst_dir = args.dst_dir or source_dataset.data_path
    dst_dir = (
        dst_dir if dst_dir else osp.join(args.project_dir, list(project.working_tree.sources)[0])
    )
    result.save(dst_dir, save_media=True)

    log.info("Results have been saved to '%s'" % dst_dir)



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import threading
from contextlib import ExitStack, contextmanager
from functools import partial, wraps
from typing import Any, Callable, ContextManager, Dict, Optional, Tuple, TypeVar

from attrs import frozen

from datumaro.util import optional_arg_decorator

T = TypeVar("T")


class Scope:
    """
    A context manager that allows to register error and exit callbacks.
    """

    _thread_locals = threading.local()

    @frozen
    class _ExitHandler:
        callback: Callable[[], Any]
        ignore_errors: bool = True

        def __exit__(self, exc_type, exc_value, exc_traceback):
            try:
                self.callback()
            except Exception:
                if not self.ignore_errors:
                    raise

    @frozen
    class _ErrorHandler(_ExitHandler):
        def __exit__(self, exc_type, exc_value, exc_traceback):
            if exc_type:
                return super().__exit__(
                    exc_type=exc_type, exc_value=exc_value, exc_traceback=exc_traceback
                )

    def __init__(self):
        self._stack = ExitStack()
        self.enabled = True

    def on_error_do(
        self,
        callback: Callable,
        *args,
        kwargs: Optional[Dict[str, Any]] = None,
        ignore_errors: bool = False,
    ):
        """
        Registers a function to be called on scope exit because of an error.

        If ignore_errors is True, the errors from this function call
        will be ignored.
        """

        self._register_callback(
            self._ErrorHandler,
            ignore_errors=ignore_errors,
            callback=callback,
            args=args,
            kwargs=kwargs,
        )

    def on_exit_do(
        self,
        callback: Callable,
        *args,
        kwargs: Optional[Dict[str, Any]] = None,
        ignore_errors: bool = False,
    ):
        """
        Registers a function to be called on scope exit.
        """

        self._register_callback(
            self._ExitHandler,
            ignore_errors=ignore_errors,
            callback=callback,
            args=args,
            kwargs=kwargs,
        )

    def _register_callback(
        self,
        handler_type,
        callback: Callable,
        args: Tuple[Any] = None,
        kwargs: Dict[str, Any] = None,
        ignore_errors: bool = False,
    ):
        if args or kwargs:
            callback = partial(callback, *args, **(kwargs or {}))

        self._stack.push(handler_type(callback, ignore_errors=ignore_errors))

    def add(self, cm: ContextManager[T]) -> T:
        """
        Enters a context manager and adds it to the exit stack.

        Returns: cm.__enter__() result
        """

        return self._stack.enter_context(cm)

    def enable(self):
        self.enabled = True

    def disable(self):
        self.enabled = False

    def close(self):
        self.__exit__(None, None, None)

    def __enter__(self) -> Scope:
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not self.enabled:
            return

        self._stack.__exit__(exc_type, exc_value, exc_traceback)
        self._stack.pop_all()  # prevent issues on repetitive calls

    @classmethod
    def current(cls) -> Scope:
        return cls._thread_locals.current

    @contextmanager
    def as_current(self):
        previous = getattr(self._thread_locals, "current", None)
        self._thread_locals.current = self
        try:
            yield
        finally:
            self._thread_locals.current = previous


@optional_arg_decorator
def scoped(func, arg_name=None):
    """
    A function decorator, which allows to do actions with the current scope,
    such as registering error and exit callbacks and context managers.
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        with Scope() as scope:
            if arg_name is None:
                with scope.as_current():
                    ret_val = func(*args, **kwargs)
            else:
                kwargs[arg_name] = scope
                ret_val = func(*args, **kwargs)
            return ret_val

    return wrapped_func


# Shorthands for common cases
def on_error_do(callback, *args, ignore_errors=False, kwargs=None):
    return Scope.current().on_error_do(callback, *args, ignore_errors=ignore_errors, kwargs=kwargs)


def on_exit_do(callback, *args, ignore_errors=False, kwargs=None):
    return Scope.current().on_exit_do(callback, *args, ignore_errors=ignore_errors, kwargs=kwargs)


def scope_add(cm: ContextManager[T]) -> T:
    return Scope.current().add(cm)



# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp
from enum import Enum, auto

from datumaro.components.comparator import DistanceComparator, EqualityComparator, TableComparator
from datumaro.components.errors import ProjectNotFoundError
from datumaro.util.os_util import rmtree
from datumaro.util.scope import on_error_do, scope_add, scoped

from ..util import MultilineFormatter
from ..util.compare import DistanceCompareVisualizer
from ..util.errors import CliException
from ..util.project import generate_next_file_name, load_project, parse_full_revpath


class ComparisonMethod(Enum):
    table = auto()
    equality = auto()
    distance = auto()


eq_default_if = ["id", "group"]  # avoid https://bugs.python.org/issue16399


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Compares two datasets",
        description="""
        Compares two datasets. This command has multiple forms:|n
        1) %(prog)s <revpath>|n
        2) %(prog)s <revpath> <revpath>|n
        |n
        1 - Compares the current project's main target ('project')
        in the working tree with the specified dataset.|n
        2 - Compares two specified datasets.|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        Both forms use the -p/--project as a context for plugins. It can be
        useful for dataset paths in targets. When not specified, the current
        project's working tree is used.|n
        |n
        Annotations can be matched 3 ways:|n
        - by comparision table|n
        - by equality checking|n
        - by distance computation|n
        |n
        Examples:|n
        - Compare two projects by distance, match boxes if IoU > 0.7,|n
        |s|s|s|ssave results to Tensorboard:|n
        |s|s%(prog)s other/project -o diff/ -f tensorboard --iou-thresh 0.7|n
        |n
        - Compare two projects for equality, exclude annotation groups |n
        |s|s|s|sand the 'is_crowd' attribute from comparison:|n
        |s|s%(prog)s other/project/ -if group -ia is_crowd -m equality|n
        |n
        - Compare two datasets, specify formats:|n
        |s|s%(prog)s path/to/dataset1:voc path/to/dataset2:coco|n
        |n
        - Compare the current working tree and a dataset:|n
        |s|s%(prog)s path/to/dataset2:coco|n
        |n
        - Compare a source from a previous revision and a dataset:|n
        |s|s%(prog)s HEAD~2:source-2 path/to/dataset2:yolo
        """,
        formatter_class=MultilineFormatter,
    )

    formats = ", ".join(f.name for f in DistanceCompareVisualizer.OutputFormat)
    comp_methods = ", ".join(m.name for m in ComparisonMethod)

    def _parse_output_format(s):
        try:
            return DistanceCompareVisualizer.OutputFormat[s.lower()]
        except KeyError:
            raise argparse.ArgumentError(
                "format",
                message="Unknown output " "format '%s', the only available are: %s" % (s, formats),
            )

    def _parse_comparison_method(s):
        try:
            return ComparisonMethod[s.lower()]
        except KeyError:
            raise argparse.ArgumentError(
                "method",
                message="Unknown comparison "
                "method '%s', the only available are: %s" % (s, comp_methods),
            )

    parser.add_argument("first_target", help="The first dataset revpath to be compared")
    parser.add_argument(
        "second_target", nargs="?", help="The second dataset revpath to be compared"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        default=None,
        help="Directory to save comparison results " "(default: generate automatically)",
    )
    parser.add_argument(
        "-m",
        "--method",
        type=_parse_comparison_method,
        default=ComparisonMethod.table.name,
        help="Comparison method, one of {} (default: %(default)s)".format(comp_methods),
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the current project (default: current dir)",
    )
    parser.set_defaults(command=compare_command)

    distance_parser = parser.add_argument_group("Distance comparison options")
    distance_parser.add_argument(
        "--iou-thresh",
        default=0.5,
        type=float,
        help="IoU match threshold for shapes (default: %(default)s)",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=_parse_output_format,
        default=DistanceCompareVisualizer.DEFAULT_FORMAT.name,
        help="Output format, one of {} (default: %(default)s)".format(formats),
    )

    equality_parser = parser.add_argument_group("Equality comparison options")
    equality_parser.add_argument(
        "-iia", "--ignore-item-attr", action="append", help="Ignore item attribute (repeatable)"
    )
    equality_parser.add_argument(
        "-ia", "--ignore-attr", action="append", help="Ignore annotation attribute (repeatable)"
    )
    equality_parser.add_argument(
        "-if",
        "--ignore-field",
        action="append",
        help="Ignore annotation field (repeatable, default: %s)" % eq_default_if,
    )
    equality_parser.add_argument(
        "--match-images",
        action="store_true",
        help="Match dataset items by image pixels instead of ids",
    )
    equality_parser.add_argument("--all", action="store_true", help="Include matches in the output")

    return parser


def get_sensitive_args():
    return {
        compare_command: [
            "first_target",
            "second_target",
            "dst_dir",
            "project_dir",
        ],
    }


@scoped
def compare_command(args):
    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name("compare")
    dst_dir = osp.abspath(dst_dir)

    if not osp.exists(dst_dir):
        on_error_do(rmtree, dst_dir, ignore_errors=True)
        os.makedirs(dst_dir)

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    try:
        if not args.second_target:
            first_dataset = project.working_tree.make_dataset()
            second_dataset, target_project = parse_full_revpath(args.first_target, project)
            if target_project:
                scope_add(target_project)
        else:
            first_dataset, target_project = parse_full_revpath(args.first_target, project)
            if target_project:
                scope_add(target_project)

            second_dataset, target_project = parse_full_revpath(args.second_target, project)
            if target_project:
                scope_add(target_project)
    except Exception as e:
        raise CliException(str(e))

    if args.method is ComparisonMethod.table:
        comparator = TableComparator()
        (
            high_level_table,
            mid_level_table,
            low_level_table,
            comparison_dict,
        ) = comparator.compare_datasets(first_dataset, second_dataset)
        if args.dst_dir:
            comparator.save_compare_report(
                high_level_table, mid_level_table, low_level_table, comparison_dict, args.dst_dir
            )

    elif args.method is ComparisonMethod.equality:
        if args.ignore_field:
            args.ignore_field = eq_default_if

        comparator = EqualityComparator(
            match_images=args.match_images,
            ignored_fields=args.ignore_field,
            ignored_attrs=args.ignore_attr,
            ignored_item_attrs=args.ignore_item_attr,
            all=args.all,
        )
        output = comparator.compare_datasets(first_dataset, second_dataset)
        if args.dst_dir:
            comparator.save_compare_report(output, args.dst_dir)

    elif args.method is ComparisonMethod.distance:
        comparator = DistanceComparator(iou_threshold=args.iou_thresh)
        with DistanceCompareVisualizer(
            save_dir=dst_dir, comparator=comparator, output_format=args.format
        ) as visualizer:
            log.info("Saving compare to '%s'" % dst_dir)
            visualizer.save(first_dataset, second_dataset)

    return 0



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

# pylint: disable=redefined-builtin

from . import (
    compare,
    convert,
    detect_format,
    download,
    explain,
    explore,
    filter,
    generate,
    info,
    merge,
    patch,
    prune,
    stats,
    transform,
    validate,
)
from .require_project import get_project_commands

__all__ = [
    "get_non_project_commands",
    "get_project_commands",
]


def get_non_project_commands():
    return [
        ("convert", convert, "Convert dataset between formats"),
        ("detect", detect_format, "Detect the format of a dataset"),
        ("compare", compare, "Compare datasets"),
        ("dinfo", info, "Print dataset info"),
        ("download", download, "Download a publicly available dataset"),
        ("explain", explain, "Run Explainable AI algorithm for model"),
        ("explore", explore, "Explore similar datasetitems of query"),
        ("filter", filter, "Filter dataset items"),
        ("generate", generate, "Generate synthetic dataset"),
        ("merge", merge, "Merge datasets"),
        ("patch", patch, "Update dataset from another one"),
        ("prune", prune, "Prune dataset"),
        ("stats", stats, "Compute dataset statistics"),
        ("transform", transform, "Modify dataset items"),
        ("validate", validate, "Validate dataset"),
    ]



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from typing import List

from datumaro.components.errors import DatasetImportError
from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer

from .format import Kitti3dPath


class Kitti3dImporter(Importer):
    _ANNO_EXT = ".txt"

    @classmethod
    def detect(cls, context: FormatDetectionContext) -> FormatDetectionConfidence:
        context.require_file(f"{Kitti3dPath.CALIB_DIR}/*.txt")
        cls._check_ann_file(context.require_file(f"{Kitti3dPath.LABEL_DIR}/*.txt"), context)
        return FormatDetectionConfidence.MEDIUM

    @classmethod
    def _check_ann_file(cls, fpath: str, context: FormatDetectionContext) -> bool:
        with context.probe_text_file(
            fpath, "Requirements for the annotation file of Kitti 3D format"
        ) as fp:
            for line in fp:
                fields = line.rstrip("\n").split(" ")
                if len(fields) == 15 or len(fields) == 16:
                    return True
                raise DatasetImportError(
                    f"Kitti 3D format txt file should have 15 or 16 fields for "
                    f"each line, but the read line has {len(fields)} fields: "
                    f"fields={fields}."
                )
            raise DatasetImportError("Empty file is not allowed.")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls._ANNO_EXT]

    @classmethod
    def find_sources(cls, path):
        # return [{"url": path, "format": "kitti3d"}]
        sources = cls._find_sources_recursive(
            path, "", "kitti3d", dirname=Kitti3dPath.LABEL_DIR, file_filter=lambda p: osp.isdir(p)
        )
        if len(sources) == 0:
            return [{"url": path, "format": "kitti3d"}]
        else:
            return sources



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp


def get_samples_path():
    cur_dir = osp.dirname(__file__)
    return osp.abspath(osp.join(cur_dir, "..", "plugins", "openvino_plugin", "samples"))
